/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.doc;

import java.util.List;
import java.util.ArrayList;

import freemarker.template.*;

/**
 * Implements a list of table-of-contents entries.
 *
 * @version $Id: ContentList.java 1003 2004-10-16 14:54:58Z run2000 $
 */
public final class ContentList implements TemplateListModel2, TemplateIndexedModel {
    private final List<ContentEntry> contentList;

    /** Creates a new ContentList */
    public ContentList() {
        contentList = new ArrayList<ContentEntry>();
    }

    /**
     * Retrieves an iterator to iterate over this content list.
     * @return an iterator to iterate over the current content list
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        if( contentList.isEmpty() ) {
            return null;
        }
        return new ContentListIterator( this, 0, contentList.size() - 1 );
    }

    /**
     * Release an iterator for this content list.
     * @param iterator the iterator to be released
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
    }

    /**
     * Retrieves the specified item from the list.
     *
     * @param index the index of the item to be retrieved
     * @return the specified index in the list
     * @throws TemplateModelException the specified index is out of bounds
     */
    public TemplateModel getAtIndex(long index) throws TemplateModelException {
        try {
            return contentList.get( (int)index );
        } catch (IndexOutOfBoundsException e) {
            throw new TemplateModelException(
                    "Array index out of bounds: " + Long.toString( index ), e );
        }

    }

    /**
     * Is the list empty?
     *
     * @return <code>true</code> if this object is empty, otherwise
     * <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return contentList.isEmpty();
    }

    /**
     * Adds a new entry to the list.
     *
     * @param entry the {@link ContentEntry} to be added
     */
    public void add(ContentEntry entry) {
        contentList.add( entry );
    }

    /**
     * Iterator class for ContentList. Includes special keys for finding
     * next, previous, first, and last indexes.
     */
    static class ContentListIterator extends
            FastIndexedIterator implements TemplateHashModel {

        /**
         * Constructor.
         * @param model the model to be iterated over
         * @param startIndex the beginning index of the list
         * @param endIndex the end index of the list
         */
        public ContentListIterator( TemplateIndexedModel model,
                                    long startIndex, long endIndex ) {
            super( model, startIndex, endIndex );
        }

        /**
         * Gets a {@link freemarker.template.TemplateModel} from the hash.
         *
         * @param key the name by which the <code>TemplateModel</code>
         * is identified in the template.
         * @return the <code>TemplateModel</code> referred to by the key,
         * or <code>null</code> if not found.
         */
        public TemplateModel get(String key) throws TemplateModelException {
            if( "next".equals(key)) {
                if( hasNext() ) {
                    return indexModel.getAtIndex( currentIndex );
                } else {
                    return null;
                }
            } else if( "previous".equals(key)) {
                if( currentIndex > 1 ) {
                    return indexModel.getAtIndex( currentIndex - 2 );
                } else {
                    return null;
                }
            } else if( "current".equals(key)) {
                return indexModel.getAtIndex( currentIndex - 1 );
            } else if( "first".equals(key)) {
                    return indexModel.getAtIndex( startIndex );
            } else if( "last".equals(key)) {
                return indexModel.getAtIndex( endIndex );
            } else {
                return null;
            }
        }
    }
}
