/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.doc;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

/**
 * SAX content handler that builds a {@link ContentList}. Once the list
 * has been built, the content list can be retrieved by using the
 * {@link #getContentList} method.
 *
 * @version $Id: ContentsBuilder.java 1003 2004-10-16 14:54:58Z run2000 $
 */
public final class ContentsBuilder extends DefaultHandler {
    private final ContentList m_cContents;
    private final StringBuffer m_cBuffer;
    private boolean m_bDoBuffering;
    private String m_aTitle;
    private String m_aLongTitle;
    private String m_aLink;
    private String m_aSuperTitle;
    private String m_aDescription;
    private boolean m_bGenerate;

    /** Creates new ContentsBuilder */
    public ContentsBuilder() {
        m_cContents = new ContentList();
        m_cBuffer = new StringBuffer();
        m_bDoBuffering = false;
    }

    /**
     * Event fired by the SAX parser when a start tag is encountered.
     */
    public void startElement(String uri,
                             String localName,
                             String qName,
                             Attributes attributes) throws SAXException {

        if( "content".equals(localName)) {
            m_bDoBuffering = false;
            m_aTitle = null;
            m_aLongTitle = null;
            m_aLink = null;
            m_aDescription = null;
            return;
        }

        if( "content-group".equals(localName)) {
            String aGenerate = attributes.getValue( "generate" );
            if(( aGenerate != null ) && ( "no".equals(aGenerate))) {
                m_bGenerate = false;
            } else {
                m_bGenerate = true;
            }
            if( m_aTitle != null ) {
                m_aSuperTitle = m_aTitle;
            } else {
                m_aSuperTitle = null;
            }
            return;
        }

        if( "name".equals(localName) || "link".equals(localName)
                || "long-title".equals(localName)
                || "description".equals(localName)) {
            m_bDoBuffering = true;
            m_cBuffer.setLength( 0 );
            return;
        }
    }

    /**
     * Event fired by the SAX parser when an end tag is encountered.
     */
    public void endElement(String uri,
                           String localName,
                           String qName) throws SAXException {
        if( "name".equals(localName)) {
            m_aTitle = m_cBuffer.toString();
            m_bDoBuffering = false;
            return;
        }
        if( "link".equals(localName)) {
            m_aLink = m_cBuffer.toString();
            m_bDoBuffering = false;
            return;
        }
        if( "long-title".equals(localName)) {
            m_aLongTitle = m_cBuffer.toString();
            m_bDoBuffering = false;
            return;
        }
        if( "description".equals(localName)) {
            m_aDescription = m_cBuffer.toString();
            m_bDoBuffering = false;
            return;
        }
        if( "content".equals(localName)) {
            if(( m_aTitle != null ) && ( m_aLink != null )) {
                ContentEntry cEntry;
                if( m_aLongTitle != null ) {
                    cEntry = new ContentEntry( m_aTitle, m_aLink, m_aLongTitle );
                } else {
                    cEntry = new ContentEntry( m_aTitle, m_aLink );
                }
                if( m_aSuperTitle != null ) {
                    cEntry.setSuperTitle( m_aSuperTitle );
                }
                if( m_aDescription != null ) {
                    cEntry.setDescription( m_aDescription );
                }
                if( m_bGenerate ) {
                    m_cContents.add( cEntry );
                }
            }
        }
        if( "content-group".equals(localName)) {
            m_bDoBuffering = false;
            m_aTitle = null;
            m_aLink = null;
            m_aLongTitle = null;
            m_aSuperTitle = null;
        }
    }

    /**
     * Event fired by the SAX parser when character data is encountered.
     */
    public void characters(char[] ch,
                           int start,
                           int length) throws SAXException {
        if( m_bDoBuffering == true ) {
            m_cBuffer.append( ch, start, length );
        }
    }

    /**
     * Returns the built content list once the SAX parse is complete.
     *
     * @return the built {@link ContentList}
     */
    public ContentList getContentList() {
        return m_cContents;
    }
}
