/*
 * ====================================================================
 *
 * Copyright (c) 2000 Attila Szegedi.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes the Expose library
 *        (http://www.szegedi.org/expose)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Expose" and "Attila Szegedi" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, please contact expose@szegedi.org.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.util.*;

/**
 * <p>A special case of {@link ObjectModel} that supports the
 * <code>collection.iterator</code> syntax to obtain a {@link freemarker.template.TemplateListModel2}
 * instance backed by this collection.</p>
 *
 * <p>The class itself also implements {@link freemarker.template.TemplateListModel2}
 * directly, since this model preserves thread safety.</p>
 *
 * @author Attila Szegedi, attila@szegedi.org
 * @version $Id: CollectionModel.java 1163 2005-10-09 10:47:15Z run2000 $
 * @deprecated this class is deprecated in favour of the classes in the
 * {@link freemarker.ext.beans2} package
 */
public class CollectionModel extends ObjectModel implements TemplateListModel2
{
	private static final String ITERATOR = "iterator";

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = -862932221330352684L;

	/**
	 * Creates a new model that wraps the specified collection object.
	 * @param collection the collection object to wrap into a model.
	 */
	public CollectionModel(Collection collection)
	{
		super(collection);
	}

	/**
	 * Returns a model wrapping the specified collection object. If there is already
	 * a cached model instance for this collection, returns the cached model instance.
	 * Models are cached using <code>WeakReference</code> objects. The caching can be turned
	 * off by setting the <kbd>expose.reflection.nocache</kbd> system property to
	 * true. In this case calling this method is equivalent to constructing a new model.
	 * @param object the collection to wrap into a model.
	 * @return the model for the collection
	 */
	public static final CollectionModel getInstance(Collection object)
	{
		if(noCache)
			return new CollectionModel(object);

		CollectionModel model = (CollectionModel)lookup(object);
		if(model == null)
		{
			model = new CollectionModel(object);
			register(model, object);
		}
		return model;
	}

	public @Override TemplateModel get(String key)
	throws
		TemplateModelException
	{
		return ITERATOR.equals(key) ? new IteratorModel((Collection)object) : super.get(key);
	}

	/**
	 * Returns the type of this object (which is TYPE_COLLECTION)
	 */
	public @Override ModelType getType()
	{
		return ModelType.TYPE_COLLECTION;
	}

	/**
	 * Returns true if the underlying collection contains no elements.
	 */
	public @Override boolean isEmpty()
	{
		return ((Collection)object).isEmpty();
	}

    /**
     * Retrive an iterator for this <code>Collection</code>.
     *
     * @return a <code>TemplateIteratorModel</code> wrapping the underlying
     * <code>java.util.Iterator</code> implementation
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        return new IteratorModelImpl(((Collection)object).iterator());
    }

    /**
     * Release the iterator once we've finished with it
     * @param iterator the iterator model to be released
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
        // We don't cache iterators, so just return
    }
}
