/*
 * ====================================================================
 *
 * Copyright (c) 2000 Attila Szegedi.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes the Expose library
 *        (http://www.szegedi.org/expose)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Expose" and "Attila Szegedi" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, please contact expose@szegedi.org.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package freemarker.ext.beans;

import freemarker.template.*;

import java.beans.*;
import java.lang.reflect.*;
import java.util.*;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.math.BigDecimal;

/**
 * A class that will wrap an arbitrary object into
 * {@link TemplateHashModel} interface. It uses Beans <code>Introspector</code>
 * to dynamically discover the properties and methods. Properties will be wrapped
 * into {@link MethodModel} and {@link ScalarModel} class
 * instances. The models are cached (meaning requesting a model for same object
 * twice will return the same model instance) unless the system property
 * <kbd>expose.reflection.nocache</kbd> is set to true.
 *
 * @author Attila Szegedi, attila@szegedi.org
 * @version $Id: ObjectModel.java 1163 2005-10-09 10:47:15Z run2000 $
 * @deprecated this class is deprecated in favour of the classes in the
 * {@link freemarker.ext.beans2} package
 */

public class ObjectModel extends ObjectModelBase implements TemplateHashModel
{
    private static final Integer GENERIC_GET_KEY = 0;
    private static final String GENERIC_GET = "get";
    private static final Class[] STRING_PARAMETER = new Class[] { java.lang.String.class };
    private static final Class[] OBJECT_PARAMETER = new Class[] { java.lang.Object.class };

    // Cache of hash maps that contain already discovered properties and methods
    // for a specified class. Each key is a Class, each value is a hash map. In
    // that hash map, each key is a property/method name, each value is a
    // MethodDescriptor or a PropertyDescriptor assigned to that property/method.
    private static final Map<Class, Object> classCache = new HashMap<Class, Object>();
    private static Set<String> cachedClassNames = new HashSet<String>();

    // Cached template models that implement member properties and methods for this
    // instance. Keys are FeatureDescriptor instances (from classCache values),
    // values are either ReflectionMethodModels/ReflectionScalarModels
    private transient Map<FeatureDescriptor, TemplateModel> memberMap = new HashMap<FeatureDescriptor, TemplateModel>();

    /** Serialization UUID for this class. */
    private  static final long serialVersionUID = -7937617645779824364L;

    /**
     * Creates a new model that wraps the specified object.
     * Note that there are specialized subclasses of this class for wrapping arrays,
     * collections, enumeration, iterators, and maps. Note also that the superclass can
     * be used to wrap String objects if only scalar functionality is needed.
     * You can also choose to delegate the choice over which model class is used for
     * wrapping to {@link BeansWrapper#wrap(Object)}.
     * @param object the object to wrap into a model.
     */
    public ObjectModel(Object object)
    {
        super(object);
        Class clazz = object.getClass();
        synchronized(classCache)
        {
            if(!classCache.containsKey(clazz))
            {
                String className = clazz.getName();
                if(cachedClassNames.contains(className))
                {
                    // Class reload detected, throw away caches
                    classCache.clear();
                    cachedClassNames = new HashSet<String>();
                    ObjectModelBase.clearModelCache();
                }
                classCache.put(clazz, populateClassMap(clazz));
                cachedClassNames.add(className);
            }
        }
    }

    /**
     * Returns a model wrapping the specified object. If there is already
     * a cached model instance for this object, returns the cached model instance.
     * Models are cached using <code>WeakReference</code> objects. The caching can be turned
     * off by setting the <kbd>expose.reflection.nocache</kbd> system property to
     * true. In this case calling this method is equivalent to constructing a new model.
     * Note that there are specialized subclasses of this class for wrapping arrays,
     * collections, enumeration, iterators, and maps. Note also that the superclass can
     * be used to wrap String objects if only scalar functionality is needed.
     * You can also choose to delegate the choice over which model class is used for
     * wrapping to {@link BeansWrapper#wrap(Object)}.
     * @param object the object to wrap into a model.
     * @return the model for the object.
     */
    public static final ObjectModel getInstance(Object object)
    {
        if(noCache)
            return new ObjectModel(object);

        ObjectModel model = (ObjectModel)lookup(object);
        if(model == null)
        {
            model = new ObjectModel(object);
            register(model, object);
        }
        return model;
    }

    /**
     * Returns the type of this object (which is TYPE_OBJECT)
     */
    public @Override ModelType getType()
    {
        return ModelType.TYPE_OBJECT;
    }

    /**
     * Uses Beans introspection to locate a property or method with name matching the
     * key name. If a method or property is found, it is wrapped into
     * {@link MethodModel} or {@link ScalarModel} instance and
     * returned. Models for various properties and methods are cached on a per-class
     * basis, so the costly introspection is performed only once per property or method
     * of a class. (Sidenote: this also implies that any class whose method has been
     * called will be strongly referred to by the framework and will not become
     * unloadable until this class has been unloaded first. Normally this is not an
     * issue, but can be in a rare scenario where you create many classes on the fly.
     * Also, as the cache grows with new classes and methods introduced to the
     * framework, it may appear as if it were leaking memory.)
     * If no method or propery matching the key is found, the framework
     * will try to invoke methods with signature <code>get(java.lang.String)</code>,
     * then <code>get(java.lang.Object)</code>.
     * @throws TemplateModelException if there was no property nor method nor
     * a generic <code>get</code> method to invoke.
     */
    public TemplateModel get(String key)
        throws
        TemplateModelException
    {
        Map keyMap;
        Class clazz = object.getClass();
        synchronized(classCache)
        {
            keyMap = (Map)classCache.get(clazz);
        }

        try
        {
            FeatureDescriptor fd = (FeatureDescriptor)keyMap.get(key);

            return
                fd == null
                ? invokeGenericGet(keyMap, key)
                : invokeThroughDescriptor(fd);
        }
        catch(TemplateModelException e)
        {
            throw e;
        }
        catch(Exception e)
        {
            java.io.StringWriter sw = new java.io.StringWriter();
            java.io.PrintWriter pw = new java.io.PrintWriter(sw);
            pw.println("Failed on " + object.getClass().getName());
            e.printStackTrace(pw);
            throw new TemplateModelException(sw.toString());
        }
    }

    /**
     * Populates a map with property and method descriptors for a specified
     * class. If any property or method descriptors specifies a read method
     * that is not accessible, replaces it with appropriate accessible method
     * from a superclass or interface. Introspection exceptions are dumped to
     * System.err and otherwise ignored, as there is no standard logging
     * facility in FM-Classic.
     */
    private static Map<Object, Object> populateClassMap(Class clazz)
    {
        // Populate first from bean info
        Map<Object, Object> map = populateClassMapWithBeanInfo(clazz);
        // Next add the generic "get" method
        Method genericGet = findGenericGetMethod(clazz);
        if(genericGet != null)
        {
            if(map.isEmpty())
                map = Collections.<Object,Object>singletonMap(GENERIC_GET_KEY, genericGet);
            else
                map.put(GENERIC_GET_KEY, genericGet);
        }
        return map;
    }

    private static Map<Object, Object> populateClassMapWithBeanInfo(Class clazz)
    {
        try
        {
            BeanInfo beanInfo = Introspector.getBeanInfo(clazz);
            PropertyDescriptor[] pda = beanInfo.getPropertyDescriptors();
            MethodDescriptor[] mda = beanInfo.getMethodDescriptors();

            int memberCount = pda.length + mda.length;
            Map<Object, Object> classMap = new HashMap<Object, Object>(memberCount * 4/3, .75f);

            for(int i = pda.length - 1; i >= 0; --i)
            {
                PropertyDescriptor pd = pda[i];
                if(pd instanceof IndexedPropertyDescriptor)
                {
                    IndexedPropertyDescriptor ipd = (IndexedPropertyDescriptor)pd;
                    Method readMethod = ipd.getReadMethod();
                    Method publicReadMethod = BeansWrapper.getPublicMethodOrNull(readMethod);
                    if(publicReadMethod != null)
                    {
                        try
                        {
                            if(readMethod != publicReadMethod)
                                ipd.setIndexedReadMethod(publicReadMethod);
                            classMap.put(ipd.getName(), ipd);
                        }
                        catch(IntrospectionException e)
                        {
                            e.printStackTrace();
                        }
                    }
                }
                else
                {
                    Method readMethod = pd.getReadMethod();
                    Method publicReadMethod = BeansWrapper.getPublicMethodOrNull(readMethod);
                    if(publicReadMethod != null)
                    {
                        try
                        {
                            if(readMethod != publicReadMethod)
                                pd.setReadMethod(publicReadMethod);
                            classMap.put(pd.getName(), pd);
                        }
                        catch(IntrospectionException e)
                        {
                            e.printStackTrace();
                        }
                    }
                }
            }

            for(int i = mda.length - 1; i >= 0; --i)
            {
                MethodDescriptor md = mda[i];
                Method readMethod = md.getMethod();
                Method publicReadMethod = BeansWrapper.getPublicMethodOrNull(readMethod);
                if(publicReadMethod != null)
                {
                    if(readMethod != publicReadMethod)
                    {
                        MethodDescriptor newMd = new MethodDescriptor(publicReadMethod, md.getParameterDescriptors());
                        copyMethodDescriptor(md, newMd);
                        md = newMd;
                    }
                    classMap.put(md.getName(), md);
                }
            }
            return classMap;
        }
        catch(IntrospectionException e)
        {
            e.printStackTrace();
            return Collections.emptyMap();
        }
    }

    private static void copyMethodDescriptor(MethodDescriptor md, MethodDescriptor newMd)
    {
        newMd.setDisplayName(md.getDisplayName());
        newMd.setExpert(md.isExpert());
        newMd.setHidden(md.isHidden());
        newMd.setName(md.getName());
        newMd.setPreferred(md.isPreferred());
        newMd.setShortDescription(md.getShortDescription());
        Enumeration attributeNames = md.attributeNames();
        while(attributeNames.hasMoreElements())
        {
            String attributeName = (String)attributeNames.nextElement();
            newMd.setValue(attributeName, md.getValue(attributeName));
        }
    }

    private static Method findGenericGetMethod(Class clazz)
    {
        Method genericGet = null;
        try
        {
            // First try "get(java.lang.String)"
            genericGet = BeansWrapper.getPublicMethod(clazz.getDeclaredMethod(GENERIC_GET, STRING_PARAMETER));
        }
        catch(NoSuchMethodException e)
        {
            // Not found...
        }

        if(genericGet == null)
        {
            try
            {
                // Next try "get(java.lang.Object)"
                genericGet = BeansWrapper.getPublicMethodOrNull(clazz.getDeclaredMethod(GENERIC_GET, OBJECT_PARAMETER));
            }
            catch(NoSuchMethodException e)
            {
                // Not found...
            }
        }
        return genericGet;
    }

    private TemplateModel invokeThroughDescriptor(FeatureDescriptor fd)
        throws
        NoSuchMethodException,
        IllegalAccessException,
        InvocationTargetException
    {
        // See if this particular instance has a cached implementation
        // for the requested feature descriptor
        TemplateModel member;
        synchronized(memberMap)
        {
            member = memberMap.get(fd);
        }

        if(member != null)
            return member;

        TemplateModel retval;
        if(fd instanceof IndexedPropertyDescriptor)
        {
            retval = member = new MethodModel(object, ((IndexedPropertyDescriptor)fd).getIndexedReadMethod());
        }
        else if(fd instanceof PropertyDescriptor)
        {
            PropertyDescriptor pd = (PropertyDescriptor)fd;
            Class clazz = pd.getPropertyType();
            Method readMethod = pd.getReadMethod();

            if(BeansWrapper.isScalar(clazz))
            {
                retval = member = getScalarModel( readMethod, clazz );
            }
            else
            {
                retval = BeansWrapper.wrap(readMethod.invoke(object));
                // (member == null) condition remains, as we don't cache these
            }
        }
        else // if(fd instanceof MethodDescriptor)
        {
            retval = member = new MethodModel(object, ((MethodDescriptor)fd).getMethod());
        }
        // If new cacheable member was created, cache it
        if(member != null)
        {
            synchronized(memberMap)
            {
                memberMap.put(fd, member);
            }
        }
        return retval;
    }

    private ScalarModel getScalarModel( Method readMethod, Class clazz ) {

        if( clazz.isPrimitive() ) {
            if(( clazz == Byte.TYPE ) || ( clazz == Short.TYPE ) ||
              ( clazz == Integer.TYPE ) || ( clazz == Long.TYPE ) ||
              ( clazz == Float.TYPE ) || ( clazz == Double.TYPE )) {
                return new NumberModel(object, readMethod);
            }
        }
        if( Number.class.isAssignableFrom( clazz )) {
            if(( clazz == Byte.class ) || ( clazz == Short.class ) ||
              ( clazz == Integer.class ) || ( clazz == Long.class ) ||
              ( clazz == Float.class ) || ( clazz == Double.class ) ||
              ( clazz == BigInteger.class ) || ( clazz == BigDecimal.class )) {
                return new NumberModel(object, readMethod);
            }
        }

        return new ScalarModel(object, readMethod);
    }

    private TemplateModel invokeGenericGet(Map keyMap, String key)
        throws
        IllegalAccessException,
        InvocationTargetException,
        TemplateModelException
    {
        Method genericGet = (Method)keyMap.get(GENERIC_GET_KEY);
        if(genericGet == null)
            throw new TemplateModelException("No such key: " + key + " on instance of " + object.getClass().getName());

        return BeansWrapper.wrap(genericGet.invoke(object, new Object[] { key }));
    }

    /**
     * Serialized form only contains the wrapped object itself. We need to
     * recreate the transient values from scratch.
     */
    private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
        stream.defaultReadObject();
        memberMap = new HashMap<FeatureDescriptor, TemplateModel>();
    }
}
