/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.IOException;

import freemarker.template.*;

/**
 * @author Nicholas Cull
 * @version $Id: ArrayWrapper.java 1164 2005-10-09 11:57:04Z run2000 $
 * @since 1.9
 */
public class ArrayWrapper implements ObjectWrapper, TemplateListModel2,
        TemplateWriteableIndexedModel, TemplateHashModel,
        TemplateObjectModel, TemplateScalarModel, Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = -3705819058406074301L;

    private Object object;
    private transient int length;

    public ArrayWrapper() {
        length = 0;
    }

    public ArrayWrapper( Object object ) {
        this.object = object;
        length = Array.getLength( object );
    }

    public void setObject( Object object ) {
        this.object = object;
        length = Array.getLength( object );
    }

    public boolean isEmpty() throws TemplateModelException {
        return object == null;
    }

    public TemplateModel get( String key ) throws TemplateModelException {
        if( "length".equals( key )) {
            return new FastNumber( length );
        }
        return null;
    }

    public TemplateIteratorModel templateIterator() throws TemplateModelException {

        return new FastIndexedWriteableIterator( this, 0, length - 1 );
    }

    public void releaseIterator( TemplateIteratorModel iterator ) {
        // Do nothing
    }

    public TemplateModel getAtIndex( long index ) throws TemplateModelException {

        if( index < 0 || index >= length ) {
            throw new TemplateModelException( "Array index out of bounds" );
        }

        Object result;
        try {
            result = Array.get( object, (int)index );
        } catch( IllegalArgumentException e ) {
            throw new TemplateModelException( "Cannot get array at index: " +
                    e.getMessage(), e );
        } catch( ArrayIndexOutOfBoundsException e ) {
            throw new TemplateModelException( "Cannot get array at index: " +
                    e.getMessage(), e );
        }
        return Wrapper.wrap( result );
    }

    /**
     * Set the value corresponding to the given index.
     *
     * @param index the index of the underlying value we're interested in
     * @param model the model to be added to the list
     * @throws freemarker.template.TemplateModelException
     *          the value could not be determined, possibly
     *          due to an index out-of-bounds, or an otherwise undefined value
     */
    public void putAtIndex( long index, TemplateModel model ) throws TemplateModelException {

        if( index < 0 || index >= length ) {
            throw new TemplateModelException( "Array index out of bounds" );
        }

        Class<?> type = object.getClass().getComponentType();
        if( type == null ) {
            throw new TemplateModelException( "Cannot set array at index: no component type returned" );
        }

        try {
            Object value = Unwrapper.unwrap( model, type );
            Array.set( object, (int)index, value );
        } catch( IllegalArgumentException e ) {
            throw new TemplateModelException( "Cannot set array at index: " +
                    e.getMessage(), e );
        } catch( ArrayIndexOutOfBoundsException e ) {
            throw new TemplateModelException( "Cannot set array at index: " +
                    e.getMessage(), e );
        } catch( TemplateModelException e ) {
            throw new TemplateModelException( "Cannot set array at index: " +
                    e.getMessage(), e );
        } catch( ClassCastException e ) {
            throw new TemplateModelException( "Cannot set array at index: " +
                    e.getMessage(), e );
        } catch( NullPointerException e ) {
            throw new TemplateModelException( "Cannot set array at index: " +
                    e.getMessage(), e );
        }
    }

    /**
     * Return the underlying object to the reflection mechanism. Any variables,
     * methods or properties can be called directly via reflection.
     *
     * @return the underlying object for this template model
     * @throws freemarker.template.TemplateModelException
     *          the object could not be returned
     */
    public Object getAsObject() throws TemplateModelException {
        return object;
    }

    /**
     * Returns the array's value as a <code>String</code>.
     *
     * @return the <code>String</code> value of this scalar.
     */
    public String getAsString() throws TemplateModelException {
        return Arrays.asList( object ).toString();
    }

    /**
     * Creates and returns a copy of this object.
     *
     * @return a clone of this instance.
     * @throws RuntimeException the clone operation was not successful
     * @throws OutOfMemoryError if there is not enough memory.
     * @see Cloneable
     */
    public Object clone() {
        try {
            return super.clone();
        } catch( CloneNotSupportedException e ) {
            throw new RuntimeException( "Clone method threw an exception: " + e.getMessage() );
        }
    }

    /**
     * Recreate the length field when deserializing the underlying array.
     */
    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {

        stream.defaultReadObject();
        length = Array.getLength( object );
    }
}
