/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.beans.*;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Field;

/**
 * A simple test class for testing introspection by Bean or Reflection packages.
 *
 * @author Nicholas Cull
 * @version $Id: BeanExplorer.java 1109 2005-09-09 11:21:12Z run2000 $
 */
public class BeanExplorer {

    // Dummy fields for testing against the introspector
    public int test1;
    protected int test2;
    int test3;
    private int test4;
    public static final int test5 = 4;

    private static void printModifiers( int modifier ) {
        System.out.print( "Modifiers: " );
        if( Modifier.isPublic( modifier )) {
            System.out.print( "Public " );
        } else if( Modifier.isProtected( modifier )) {
            System.out.print( "Protected " );
        } else if( Modifier.isPrivate( modifier )) {
            System.out.println( "Private " );
        } else {
            System.out.println( "Package-protected (?) " );
        }

        if( Modifier.isStatic( modifier )) {
            System.out.print( "Static " );
        } else if( Modifier.isAbstract( modifier )) {
            System.out.print( "Abstract " );
        } else {
            System.out.print( "Dynamic " );
        }

        if( Modifier.isSynchronized( modifier )) {
            System.out.println( "Synchronized " );
        }

        if( Modifier.isFinal( modifier )) {
            System.out.print( "Final " );
        }
        System.out.println();
    }

    public static void main( String[] args ) throws Exception {
        String className = args[0];
        Class c1 = Class.forName( className );
/*
        Perl5Substitute p5s = Perl5Substitute.getInstance();
        List arguments = new ArrayList();
        arguments.add( "s/boo/foo/" );
        TemplateModel result = p5s.exec( arguments );
        Class c1 = result.getClass();
*/

        BeanInfo beanInfo = Introspector.getBeanInfo( c1 );

        System.out.println( "Class " + className + " found, introspection follows..." );
        printModifiers( c1.getModifiers() );

        System.out.println( "Properties:" );
        PropertyDescriptor[] pds = beanInfo.getPropertyDescriptors();

        for( int i = 0; i < pds.length; i++ ) {
            PropertyDescriptor pd = pds[ i ];
            System.out.println( "Descriptor " + Integer.toString( i + 1 ) + ":" );
            System.out.println( "Property name is: " + pd.getName() );
            System.out.println( "Property type is: " + pd.getPropertyType().toString() );

            Method readMethod = pd.getReadMethod();
            if( readMethod != null ) {
                System.out.println( "Read method is: " + readMethod.toString() );
                printModifiers( readMethod.getModifiers() );
            }
            Method writeMethod = pd.getWriteMethod();
            if( writeMethod != null ) {
                System.out.println( "Write method is: " + writeMethod.toString() );
                printModifiers( writeMethod.getModifiers() );
            }
            if( pd instanceof IndexedPropertyDescriptor ) {
                System.out.println( "This is an indexed property descriptor." );
                IndexedPropertyDescriptor ipd = (IndexedPropertyDescriptor) pd;

                System.out.println( "Indexed property type is: " + ipd.getIndexedPropertyType().toString() );
                Method indexedReadMethod = ipd.getIndexedReadMethod();
                if( indexedReadMethod != null ) {
                    System.out.println( "Indexed read method is: " + indexedReadMethod.toString() );
                    printModifiers( indexedReadMethod.getModifiers() );
                }
                Method indexedWriteMethod = ipd.getIndexedWriteMethod();
                if( indexedWriteMethod != null ) {
                    System.out.println( "Indexed write method is: " + indexedWriteMethod.toString() );
                    printModifiers( indexedWriteMethod.getModifiers() );
                }
            }
            System.out.println();
        }

        System.out.println( "Methods:" );
        MethodDescriptor[] mds = beanInfo.getMethodDescriptors();

        for( int i = 0; i < mds.length; i++ ) {
            MethodDescriptor md = mds[ i ];
            System.out.println( "Descriptor " + Integer.toString( i + 1 ) + ":" );
            Method m = md.getMethod();
            System.out.println( "Method is: " + m.toString() );
            printModifiers( m.getModifiers() );
            System.out.println();
        }

        System.out.println( "Events:" );
        EventSetDescriptor[] eds = beanInfo.getEventSetDescriptors();

        for( int i = 0; i < eds.length; i++ ) {
            EventSetDescriptor ed = eds[ i ];
            System.out.println( "Descriptor " + Integer.toString( i + 1 ) + ":" );
            System.out.println( "Listener type is: " + ed.getListenerType().toString() );
            Method m = ed.getAddListenerMethod();
            System.out.println( "Add listener method: " + m.toString() );
            printModifiers( m.getModifiers() );

            m = ed.getRemoveListenerMethod();
            System.out.println( "Remove listener method: " + m.toString() );
            if( ed.isUnicast() ) {
                System.out.println( "Unicast event." );
            } else {
                System.out.println( "Multicast event." );
            }
            System.out.println();
        }

        // Now we go to the reflection API to get field information
        System.out.println( "Fields:" );
        Field[] fields = c1.getFields();

        for( int i = 0; i < fields.length; i++ ) {
            Field field = fields[ i ];
            System.out.println( "Descriptor " + Integer.toString( i + 1 ) + ":" );
            System.out.println( "Field name is: " + field.getName() );
            System.out.println( "Field type is: " + field.getType().toString() );
            printModifiers( field.getModifiers() );
            System.out.println();
        }
    }
}
