/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Enumeration;

import freemarker.template.TemplateModelException;
import freemarker.template.TemplateListModel2;
import freemarker.template.TemplateIteratorModel;

/**
 * Wraps an enumeration method.
 *
 * @author Nicholas Cull
 * @version $Id: EnumerationMethodWrapper.java 1100 2005-09-07 13:31:21Z run2000 $
 * @since 1.9
 */
public class EnumerationMethodWrapper implements TemplateListModel2 {

    protected Object object;
    protected Method method;
    protected Object[] paramValues;

    public EnumerationMethodWrapper() {
    }

    public EnumerationMethodWrapper( Method method, Object[] paramValues ) {
        this.method = method;
        this.paramValues = paramValues;
    }

    public EnumerationMethodWrapper( Method method, Object[] paramValues,
                                     Object object ) {
        this.method = method;
        this.paramValues = paramValues;
        this.object = object;
    }

    public void setObject( Object object ) {
        this.object = object;
    }

    public void setMethod( Method method ) {
        this.method = method;
    }

    public void setParamValues( Object[] paramValues ) {
        this.paramValues = paramValues;
    }

    public boolean isEmpty() throws TemplateModelException {
        return method == null;
    }

    /**
     * Retrieves an iterator to iterate over this list. Can choose to return
     * <code>null</code> if there are no elements in the list.
     *
     * @return an iterator to iterate over the current list.
     * @throws freemarker.template.TemplateModelException
     *          the next item in the list can't be
     *          retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator()
            throws TemplateModelException {

        Enumeration enumeration;

        try {
            enumeration = (Enumeration)Wrapper.invokeMethod( method, object, paramValues );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() could not be accessed", e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() threw an exception: " + e.getTargetException().getMessage(),
                    (Exception)e.getTargetException() );
        }
        if( enumeration != null ) {
            return new EnumerationWrapper<Enumeration>( enumeration );
        }
        return null;
    }

    /**
     * <p>Returns the used iterator to the list model. Implement this method
     * when you want to use an object pool of <code>TemplateIterator</code> objects.
     * Otherwise, leave the implementation of this method blank.</p>
     *
     * <p>Note that if the iterator returned in {@link #templateIterator} is
     * <code>null</code>, this method will not be called for the
     * <code>null</code> iterator.</p>
     *
     * @param iterator the iterator to be returned to the object pool, if any
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
        // Nothing to see here
    }

}
