/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import freemarker.template.TemplateIndexedModel;
import freemarker.template.TemplateWriteableIndexedModel;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateIteratorModel;

import java.util.List;
import java.util.ListIterator;
import java.io.Serializable;

/**
 * @author Nicholas Cull
 * @version $Id: ListWrapper.java 1164 2005-10-09 11:57:04Z run2000 $
 * @since 1.9
 */
public class ListWrapper<T extends List> extends CollectionWrapper<T>
        implements TemplateIndexedModel, TemplateWriteableIndexedModel,
        Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = 6763036843643470576L;

    public ListWrapper() {
    }

    public ListWrapper( T object ) {
        super( object );
    }

    /**
     * Retrieves an iterator to iterate over this collection.
     *
     * @return an iterator to iterate over the current list.
     * @throws freemarker.template.TemplateModelException
     *          the next item in the list can't be
     *          retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator()
            throws TemplateModelException {
        List coll = object;
        return new ListIteratorWrapper<ListIterator>( coll.listIterator() );
    }

    /**
     * Get the value corresponding to the given index. This corresponds
     * correspond to an index into an List such as a <code>java.util.Vector</code>.
     *
     * @param index the index of the underlying value we're interested in
     * @return a <code>TemplateModel</code> representing the value for the given index
     * @throws freemarker.template.TemplateModelException
     *          the value could not be determined, possibly
     *          due to an index out-of-bounds, or an otherwise undefined value
     */
    public TemplateModel getAtIndex( long index ) throws TemplateModelException {
        List list = object;

        try {
            return Wrapper.wrap( list.get( (int) index ));
        } catch( IndexOutOfBoundsException e ) {
            throw new TemplateModelException( "Cannot retrieve a list item with index " + index );
        }
    }

    /**
     * Set the value corresponding to the given index. Traditionally this would
     * correspond to an index into an array, or similar structure, such as a
     * <code>java.util.Vector</code>.
     *
     * @param index the index of the underlying value we're interested in
     * @param model the model to be added to the list
     * @throws freemarker.template.TemplateModelException
     *          the value could not be determined, possibly
     *          due to an index out-of-bounds, or an otherwise undefined value
     */
    public void putAtIndex( long index, TemplateModel model ) throws TemplateModelException {
        Object item = Unwrapper.defaultUnwrap( model );
        List list = object;

        try {
            list.set( (int)index, item );
        } catch( IndexOutOfBoundsException e ) {
            throw new TemplateModelException( "List index out of bounds for setter: " + index );
        } catch( UnsupportedOperationException e ) {
            throw new TemplateModelException( "List assignment is not supported by this list " );
        }
    }
}
