/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import freemarker.template.*;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Iterator;
import java.io.Serializable;

/**
 * @author Nicholas Cull
 * @version $Id: MapWrapper.java 1164 2005-10-09 11:57:04Z run2000 $
 * @since 1.9
 */
public class MapWrapper<T extends Map> extends BeanWrapper<T>
        implements TemplateMethodModel2, TemplateListModel2, Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = -4820592296667036613L;

    public MapWrapper() {
    }

    public MapWrapper( T object ) {
        super( object );
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        Map map = object;
        int size = arguments.size();

        if( size < 1 || size > 2 ) {
            throw new TemplateModelException( "Map wrapper method call must contain exactly one or two arguments" );
        }

        Object key = Unwrapper.defaultUnwrap( arguments.get(0));
        if( size == 1 ) {
            return Wrapper.wrap( map.get( key ));
        } else {
            Object value = Unwrapper.defaultUnwrap( arguments.get(1));
            map.put( key, value );
            return null;
        }
    }

    /**
     * Retrieves an iterator to iterate over this Map.
     *
     * @return an iterator to iterate over the current list.
     * @throws freemarker.template.TemplateModelException
     *          the next item in the list can't be
     *          retrieved, or no next item exists.
     */
    public TemplateIteratorModel templateIterator() throws TemplateModelException {
        Map map = object;
        Set set = map.entrySet();
        return new IteratorWrapper<Iterator>( set.iterator() );
    }

    /**
     * <p>Returns the used iterator to the list model. Implement this method
     * when you want to use an object pool of <code>TemplateIterator</code> objects.
     * Otherwise, leave the implementation of this method blank.</p>
     * <p/>
     * <p>Note that if the iterator returned in {@link #templateIterator} is
     * <code>null</code>, this method will not be called for the
     * <code>null</code> iterator.</p>
     *
     * @param iterator the iterator to be returned to the object pool, if any
     */
    public void releaseIterator( TemplateIteratorModel iterator ) {
        // Nothing to do here
    }
}
