/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.beans.IntrospectionException;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;

/**
 * Wraps an identifier. An identifier can be a method name, a property name,
 * or a field name. This class encapsulates these possibilities.
 * Note that an identifier could be an indexed property name, in which case
 * we use the IndexedPropertyWrapper, a subclass of this.
 *
 * @author Nicholas Cull
 * @version $Id: MethodWrapper.java 1109 2005-09-09 11:21:12Z run2000 $
 * @since 1.9
 */
public class MethodWrapper implements TemplateMethodModel2 {

    protected static final Object[] emptyParamValues = new Object[] { };
    protected static final Converter[] emptyConverters = new Converter[] { };
    protected static final TemplateModel[] emptyModelArguments = new TemplateModel[] { };

    protected Object object;
    protected String identifier;

    public MethodWrapper() {
    }

    public MethodWrapper( Object object, String identifier ) {
        this.object = object;
        this.identifier = identifier;
    }

    public void setObject( Object object ) {
        this.object = object;
    }

    public void setIdentifier( String identifier ) {
        this.identifier = identifier;
    }

    public boolean isEmpty() throws TemplateModelException {
        return object == null;
    }

    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        TemplateModel[] args;
        Features<?> features;

        if( arguments.size() > 0 ) {
            args = new TemplateModel[ arguments.size() ];
            arguments.toArray( args );
        } else {
            args = emptyModelArguments;
        }

        try {
            features = Wrapper.getObjectFeatures( object );
        } catch( IntrospectionException e ) {
            throw new TemplateModelException( "Could not introspect object while attempting to call method " +
                    identifier + "(): " + e.getMessage(), e );
        }

        List<Method> methods = features.getMethods( identifier );
        List<Method> candidates = getCandidates( methods, args.length );

        if( candidates.isEmpty() ) {
            throw new TemplateModelException( "Cannot find a candidate method model, wrong number of arguments" );
        }

        // Figure out which method we're trying to call here, unwrap
        // the argument list as appropriate, and call the target method
        for( Method method : candidates ) {
            Class params[] = method.getParameterTypes();

            // Match up argument types
            Converter[] converters = getConversion( params, args );

            if( converters != null ) {
                // Perform the argument conversions and call the method
                Object[] paramValues = convertParameters( converters, args );
                Class returnType = method.getReturnType();

                if( Iterator.class.isAssignableFrom( returnType )) {
                    return new IteratorMethodWrapper( method, paramValues, object );
                } else if( Enumeration.class.isAssignableFrom( returnType )) {
                    return new EnumerationMethodWrapper( method, paramValues, object );
                }
                return invokeMethod( object, method, paramValues );
            }
        }

        throw new TemplateModelException( "Could not find method call " + identifier +
                "() with correct argument types" );
    }

    private static List<Method> getCandidates( List<Method> methods, int argSize ) {
        List<Method> candidates = new ArrayList<Method>( methods.size() );

        for( Method method : methods ) {
            int length = method.getParameterTypes().length;

            if( length == argSize ) {
                candidates.add( method );
            } else if( length > argSize ) {
                break;
            }
        }
        return candidates;
    }

    private static Converter[] getConversion( Class[] params, TemplateModel[] arguments ) {

        if( params.length == 0 ) {
            return emptyConverters;
        }

        Converter[] converters = new Converter[ params.length ];
        for( int i = 0; i < params.length; i++) {
            converters[ i ] = Unwrapper.getConverter( arguments[ i ], params[ i ]);
            if( converters[ i ] == null ) {
                return null;
            }
        }
        return converters;
    }

    private static Object[] convertParameters( Converter[] converters,
            TemplateModel[] arguments ) throws TemplateModelException {

        if( converters.length == 0 ) {
            return emptyParamValues;
        }

        Object[] paramValues = new Object[ converters.length ];
        for( int i = 0; i < converters.length; i++ ) {
            paramValues[ i ] = converters[ i ].convert( arguments[ i ]);
        }
        return paramValues;
    }

    private static TemplateModel invokeMethod( Object object, Method method,
            Object[] paramValues ) throws TemplateModelException {

        try {
            Object result = Wrapper.invokeMethod( method, object, paramValues );
            return Wrapper.wrap( result );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() could not be accessed", e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() threw an exception: " + e.getTargetException().getMessage(),
                    (Exception)e.getTargetException() );
        }
    }
}
