/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.List;
import java.io.Serializable;

import freemarker.template.*;

/**
 * @author Nicholas Cull
 * @version $Id: StaticClassWrapper.java 1164 2005-10-09 11:57:04Z run2000 $
 * @since 1.9
 */
public class StaticClassWrapper<T> implements TemplateHashModel,
        TemplateWriteableHashModel, TemplateScalarModel, Serializable {

    /** Serialization UUID for this class */
    private static final long serialVersionUID = -7652511717588242325L;

    protected Class<T> clazz;

    public StaticClassWrapper() {
    }

    public StaticClassWrapper( Class<T> clazz ) {
        this.clazz = clazz;
    }

    public Class getClazz() {
        return clazz;
    }

    public void setClazz( Class<T> clazz ) {
        this.clazz = clazz;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return clazz == null;
    }

    /**
     * Gets a {@link freemarker.template.TemplateModel} from the hash.
     *
     * @param key the name by which the <code>TemplateModel</code>
     *            is identified in the template.
     * @return the <code>TemplateModel</code> referred to by the key,
     *         or <code>null</code> if not found.
     */
    public TemplateModel get( String key ) throws TemplateModelException {
        if( key == null ) {
            return null;
        }

        Features<?> features = Wrapper.getClassFeatures( clazz );
        Field field = features.getField( key );
        if( field != null ) {
            try {
                return Wrapper.wrap( field.get( null ));
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot access field " + key +
                        ": " + e.getMessage(), e );
            }
        }

        List<Method> methods = features.getMethods( key );

        if(( methods != null ) && ( ! methods.isEmpty() )) {
            return new StaticMethodWrapper( clazz, key );
        }

        throw new TemplateModelException( "Could not find class method or field " + key );
    }

    /**
     * Sets a value in the hash model.
     *
     * @param key   the hash key
     * @param model the value to be added to the hash model
     */
    public void put( String key, TemplateModel model )
            throws TemplateModelException {

        Features<?> features = Wrapper.getClassFeatures( clazz );
        Field field = features.getField( key );
        if( field != null ) {
            try {
                Class<?> type = field.getType();
                Object value = Unwrapper.unwrap( model, type );
                field.set( null, value );
                return;
            } catch( IllegalAccessException e ) {
                throw new TemplateModelException( "Cannot set field " + key +
                        ": " + e.getMessage(), e );
            }
        }

        throw new TemplateModelException( "Could not set field " + key );
    }

    /**
     * Returns the class's name as a <code>String</code>.
     *
     * @return the <code>String</code> representation of this class.
     */
    public String getAsString() throws TemplateModelException {
        return clazz.toString();
    }
}
