/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;

/**
 * @author Nicholas Cull
 * @version $Id: StaticMethodWrapper.java 1154 2005-10-09 09:22:01Z run2000 $
 * @since 1.9
 */
public class StaticMethodWrapper implements TemplateMethodModel2 {

    protected static final Object[] emptyParamValues = new Object[] { };
    protected static final Converter[] emptyConverters = new Converter[] { };
    protected static final TemplateModel[] emptyModelArguments = new TemplateModel[] { };

    protected Class clazz;
    protected String identifier;

    public StaticMethodWrapper() {
    }

    public void setClazz( Class clazz ) {
        this.clazz = clazz;
    }

    public StaticMethodWrapper( Class clazz, String identifier ) {
        this.clazz = clazz;
        this.identifier = identifier;
    }

    public Class getClazz() {
        return clazz;
    }

    public void setIdentifier( String identifier ) {
        this.identifier = identifier;
    }

    public String getIdentifier() {
        return identifier;
    }

    public boolean isEmpty() throws TemplateModelException {
        return clazz == null;
    }

    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        TemplateModel[] args;

        if( arguments.size() > 0 ) {
            args = new TemplateModel[ arguments.size() ];
            arguments.toArray( args );
        } else {
            args = emptyModelArguments;
        }

        Features features = Wrapper.getClassFeatures( clazz );
        List<Method> methods = features.getMethods( identifier );
        List<Method> candidates = getCandidates( methods, args.length );

        if( candidates.isEmpty() ) {
            throw new TemplateModelException( "Cannot find a candidate method model, wrong number of arguments" );
        }

        // Figure out which method we're trying to call here, unwrap
        // the argument list as appropriate, and call the target method
        for( Method method : candidates ) {
            Class params[] = method.getParameterTypes();

            // Match up argument types
            Converter[] converters = getConversion( params, args );
            if( converters != null ) {
                // Perform the argument conversions and call the method
                Object[] paramValues = convertParameters( converters, args );
                Class returnType = method.getReturnType();

                if( Iterator.class.isAssignableFrom( returnType )) {
                    return new IteratorMethodWrapper( method, paramValues );
                } else if( Enumeration.class.isAssignableFrom( returnType )) {
                    return new EnumerationMethodWrapper( method, paramValues );
                }
                return invokeMethod( method, paramValues );
            }
        }

        throw new TemplateModelException( "Could not find method call " + identifier +
                "() with correct argument types" );
    }

    private static List<Method> getCandidates( List<Method> methods, int argSize ) {
        List<Method> candidates = new ArrayList<Method>( methods.size() );

        for( Method method : methods ) {
            int length = method.getParameterTypes().length;

            if( length == argSize ) {
                candidates.add( method );
            } else if( length > argSize ) {
                break;
            }
        }
        return candidates;
    }

    private static Converter[] getConversion( Class params[], TemplateModel[] arguments ) {

        if( params.length == 0 ) {
            return emptyConverters;
        }

        Converter[] converters = new Converter[ params.length ];
        for( int i = 0; i < params.length; i++) {
            converters[ i ] = Unwrapper.getConverter( arguments[ i ], params[ i ]);
            if( converters[ i ] == null ) {
                return null;
            }
        }
        return converters;
    }

    private static Object[] convertParameters( Converter[] converters,
            TemplateModel[] arguments ) throws TemplateModelException {

        if( converters.length == 0 ) {
            return emptyParamValues;
        }

        Object[] paramValues = new Object[ converters.length ];
        for( int i = 0; i < converters.length; i++ ) {
            paramValues[ i ] = converters[ i ].convert( (TemplateModel) arguments[ i ]);
        }
        return paramValues;
    }

    private static TemplateModel invokeMethod( Method method,
            Object[] paramValues ) throws TemplateModelException {

        try {
            Object result = Wrapper.invokeMethod( method, null, paramValues );
            return Wrapper.wrap( result );
        } catch( IllegalAccessException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() could not be accessed", e );
        } catch( InvocationTargetException e ) {
            throw new TemplateModelException( "Method call " + method.getName() +
                    "() threw an exception", (Exception)e.getTargetException() );
        }
    }
}
