/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.beans2;

import java.util.List;
import java.util.ArrayList;

import freemarker.template.*;

/**
 * Unwraps a given FM-Classic template model back into a plain old Java object.
 * If a target class is supplied, try to convert the model using the most
 * appropriate conversion available. Otherwise, perform a default conversion
 * into a Java object.
 *
 * @author Nicholas Cull
 * @version $Id: Unwrapper.java 1158 2005-10-09 10:04:56Z run2000 $
 * @since 1.9
 */
public class Unwrapper {

    /** A list of converters, each of which maps a TemplateModel to an Object */
    private static Converter[] converters = init();

    private Unwrapper() {
        // Do nothing
    }

    private static Converter[] init() {
        List<Converter> convertList = new ArrayList<Converter>();

        // Careful to add these items in a specific order
        // Defaults come first...
        convertList.add( ConvertToObject.getInstance() );
        convertList.add( ConvertToString.getInstance() );
        convertList.add( ConvertToArrayList.getInstance() );
        convertList.add( ConvertToInteger.getInstance() );

        // Now can fill out the rest, starting with sets
        convertList.add( ConvertToHashSet.getInstance() );
        convertList.add( ConvertToTreeSet.getInstance() );

        // And Numbers
        convertList.add( ConvertToLong.getInstance() );
        convertList.add( ConvertToDouble.getInstance() );
        convertList.add( ConvertToFloat.getInstance() );
        convertList.add( ConvertToShort.getInstance() );
        convertList.add( ConvertToByte.getInstance() );
        convertList.add( ConvertToBigInteger.getInstance() );
        convertList.add( ConvertToBigDecimal.getInstance() );

        // Bool primitive
        convertList.add( ConvertToBool.getInstance() );

        // Now lists
        convertList.add( ConvertToLinkedList.getInstance() );
        convertList.add( ConvertToVector.getInstance() );

        // Finally arrays. These in turn unwrap to their component type
        convertList.add( new ConvertToObjectArray());
        convertList.add( new ConvertToArray( String.class ));
        convertList.add( new ConvertToArray( Integer.class ));
        convertList.add( new ConvertToIntArray() );
        convertList.add( new ConvertToArray( Boolean.class ));
        convertList.add( new ConvertToBoolArray() );

        // Turn the result into an array
        Converter[] converters = new Converter[ convertList.size() ];
        convertList.toArray( converters );
        return converters;
    }

    /**
     * Convert the supplied template model into a plain old Java object. If the
     * template model is a <code>TemplateObjectModel</code>, unwrap the object
     * back to the underlying object. Otherwise, perform a default conversion
     * to the nearest equivalent Java object.
     *
     * @param model the model to be unwrapped
     * @return a plain old Java object equivalent to the template model
     * @throws TemplateModelException there was a problem unwrapping the
     * template model
     */
    public static Object defaultUnwrap( TemplateModel model )
            throws TemplateModelException, ClassCastException, NullPointerException {

        return unwrap( model, Object.class );
    }

    /**
     * Convert the supplied template model into a plain old Java object of the
     * given class. Attempt to convert the model to the target class using the
     * best available conversion.
     *
     * @param model the template model to be unwrapped
     * @param target the target class for the unwrapping
     * @return a plain old Java object of the given class, equivalent to the
     * template model
     * @throws TemplateModelException the model could not be converted to the
     * target class
     */
     public static <T> T unwrap( TemplateModel model, Class<T> target )
            throws TemplateModelException, ClassCastException, NullPointerException {

        if( model == null ) {
            return null;
        }

        Converter converter = getConverter( model, target );
        if( converter != null ) {
            return (T)converter.convert( model );
        }

        throw new TemplateModelException( "Could not unwrap model: " + model.getClass().toString() );
    }

    static Converter getConverter( TemplateModel source, Class target ) {
        int size = converters.length;

        for( int i = 0; i < size; i++ ) {
            Converter converter = converters[ i ];
            if( converter.converts( source, target )) {
                return converter;
            }
        }

        return null;
    }
}
