/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.format;

import java.util.List;
import java.util.Date;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.Collections;
import java.text.SimpleDateFormat;
import java.io.Serializable;
import java.io.ObjectStreamException;
import java.lang.ref.WeakReference;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateObjectModel;
import freemarker.template.TemplateNumberModel;
import freemarker.template.FastScalar;
import freemarker.template.expression.ExpressionUtils;

/**
 * Formats numbers into the specified date format. The numbers can be
 * wrapped <code>Date</code> objects, wrapped <code>Number</code> objects, or
 * <code>TemplateNumberModel</code> values. Conversion of numbers into date
 * values is the same as for <code>java.util.Date</code> values, ie.
 * milliseconds since 1 January 1970 GMT.
 * <p>
 * Formats are specified using the same pattern as the
 * <code>java.text.SimpleDateFormat</code> class. As a performance optimization,
 * some caching of pre-parsed format templates occurs.
 * <p>
 * <b>Usage:</b><br>
 * From java:
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "formatDate", freemarker.ext.format.DateFormat.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following creates a directory path:
 * &lt;assign date = 12345678901234&gt;
 * &lt;assign output = formatDate( "dd/MM/yyyy HH:mm", date )&gt;
 * ${output}
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: DateFormat.java 1185 2005-10-13 12:24:14Z run2000 $
 * @since 1.9
 */
public class DateFormat implements TemplateMethodModel2, Serializable {

    private static final DateFormat instance = new DateFormat();

    /** Cache for pre-compiled date format objects */
    private static final Map<String, WeakReference<SimpleDateFormat>> formatCache =
            Collections.synchronizedMap(new WeakHashMap<String, WeakReference<SimpleDateFormat>>());

    /** Class UUID for serialization */
    private static final long serialVersionUID = 1202156085260321616L;

    /** Private constructor, indicating that this is a singleton. */
    private DateFormat() { }

    /**
     * Return a singleton instance of this class.
     *
     * @return a new DateFormat object
     */
    public static DateFormat getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int size = arguments.size();

        if( size != 2 ) {
            throw new TemplateModelException( "DateFormat method call must contain two arguments" );
        }

        TemplateModel keyModel = arguments.get( 0 );
        String formatTemplate = ExpressionUtils.getAsStringOrEmpty( keyModel );
        SimpleDateFormat dateFormat = getDateFormat( formatTemplate );
        TemplateModel dateModel = arguments.get( 1 );
        Object date;

        if( dateModel instanceof TemplateObjectModel ) {
            date = ((TemplateObjectModel)dateModel).getAsObject();
        } else if( dateModel instanceof TemplateNumberModel ) {
            date = new Date(((TemplateNumberModel)dateModel).getAsNumber());
        } else {
            throw new TemplateModelException( "Cannot convert target model to a Date or Number" );
        }

        if( date instanceof Date ) {
            return new FastScalar( dateFormat.format( date ));
        } else if( date instanceof Number ) {
            Date dateValue = new Date( ((Number)date).longValue() );
            return new FastScalar( dateFormat.format( dateValue ));
        } else {
            throw new TemplateModelException( "Target model is not a Date or Number" );
        }
    }

    /**
     * Handle caching of format objects. Always return a clone of the cached
     * format object, to avoid multi-threading issues.
     *
     * @param formatTemplate the format template for the value to be formatted
     * @return a SimpleDateFormat corresponding to the given format template
     */
    private static SimpleDateFormat getDateFormat( String formatTemplate ) {
        WeakReference<SimpleDateFormat> ref = formatCache.get( formatTemplate );
        SimpleDateFormat format =  null;

        if( ref != null ) {
            format = ref.get();
        }

        if( format == null ) {
            format = new SimpleDateFormat( formatTemplate );
            ref = new WeakReference<SimpleDateFormat>( format );
            formatCache.put( new String( formatTemplate ), ref );
        }

        return (SimpleDateFormat)format.clone();
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
