/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.format;

import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.Collections;
import java.text.DecimalFormat;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.lang.ref.WeakReference;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateObjectModel;
import freemarker.template.TemplateNumberModel;
import freemarker.template.FastScalar;
import freemarker.template.expression.ExpressionUtils;

/**
 * Formats numbers into the specified number format. The numbers can be
 * wrapped <code>Number</code> object or <code>TemplateNumberModel</code>
 * values.
 * <p>
 * Formats are specified using the same pattern as the
 * <code>java.text.DecimalFormat</code> class. As a performance optimization,
 * some caching of pre-parsed format templates occurs.
 * <p>
 * <b>Usage:</b><br>
 * From java:
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "formatNumber", freemarker.ext.format.NumberFormat.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following creates a directory path:
 * &lt;assign number = 12345&gt;
 * &lt;assign output = formatNumber( "###,##0.0", number )&gt;
 * ${output}
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: NumberFormat.java 1185 2005-10-13 12:24:14Z run2000 $
 * @since 1.9
 */
public class NumberFormat implements TemplateMethodModel2, Serializable {

    private static final NumberFormat instance = new NumberFormat();

    /** Cache for pre-compiled date format objects */
    private static final Map<String, WeakReference<DecimalFormat>> formatCache =
            Collections.synchronizedMap(new WeakHashMap<String, WeakReference<DecimalFormat>>());

    /** Class UUID for serialization */
    private static final long serialVersionUID = 3089437737743304579L;

    /** Private constructor, indicating that this is a singleton. */
    private NumberFormat() { }

    /**
     * Return a singleton instance of this class.
     *
     * @return a new DateFormat object
     */
    public static NumberFormat getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link freemarker.template.TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     *                  containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        int size = arguments.size();

        if( size != 2 ) {
            throw new TemplateModelException( "NumberFormat method call must contain two arguments" );
        }

        TemplateModel keyModel = arguments.get( 0 );
        String formatTemplate = ExpressionUtils.getAsStringOrEmpty( keyModel );
        TemplateModel numberModel = arguments.get( 1 );
        Object number;

        if( numberModel instanceof TemplateObjectModel ) {
            number = ((TemplateObjectModel)numberModel).getAsObject();
        } else if( numberModel instanceof TemplateNumberModel ) {
            number = new Long( ((TemplateNumberModel)numberModel).getAsNumber() );
        } else {
            throw new TemplateModelException( "Target model is not a number" );
        }

        DecimalFormat format = getDecimalFormat( formatTemplate );
        return new FastScalar( format.format( number ));
    }

    /**
     * Handle caching of format objects. Always return a clone of the cached
     * format object, to avoid multi-threading issues.
     *
     * @param formatTemplate the format template for the value to be formatted
     * @return a DecimalFormat corresponding to the given format template
     */
    private static DecimalFormat getDecimalFormat( String formatTemplate ) {
        WeakReference<DecimalFormat> ref = formatCache.get( formatTemplate );
        DecimalFormat format =  null;

        if( ref != null ) {
            format = ref.get();
        }

        if( format == null ) {
            format = new DecimalFormat( formatTemplate );
            ref = new WeakReference<DecimalFormat>( format );
            formatCache.put( new String( formatTemplate ), ref );
        }

        return (DecimalFormat)format.clone();
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
