/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.Serializable;
import java.io.ObjectStreamException;

import freemarker.template.*;

/**
 * <p>Performs a type conversion from a String model to a number model.
 * This is useful when we want to perform numeric operations on a value that
 * was introduced as a scalar value, or when we only want a number from a
 * multimodel.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "asNumber", freemarker.ext.misc.AsNumber.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * &lt;assign var1 = "5"&gt;
 * &lt;assign var2 = "3"&gt;
 * &lt;assign var3 = asNumber( var1 ) + asNumber( var2 )&gt;
 * &lt;p&gt;The sum of ${var1} and ${var2} is: ${var3}&lt;/p&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>AsNumber</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: AsNumber.java 1093 2005-08-30 12:23:57Z run2000 $
 * @since 1.8
 */
public final class AsNumber implements TemplateMethodModel2, Serializable {
    private static final AsNumber instance = new AsNumber();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 6428166403118081275L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private AsNumber() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>AsNumber</code> class
     */
    public static AsNumber getInstance() {
        return instance;
    }

    /**
     * Performs a String to number conversion using the rules that
     * <code>Long.parseLong()</code> uses.
     *
     * @param arguments a List of exactly one <code>TemplateScalarModel</code>
     * to be converted
     * @return a new <code>TemplateNumberModel</code> of the converted String
     * @throws TemplateModelException More than one argument is supplied,
     * or the model is neither a number nor a string model.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 1 ) {
            throw new TemplateModelException( "ToNumber must have exactly one argument" );
        }
        TemplateModel model = arguments.get(0);
        if( model instanceof TemplateNumberModel ) {
            if( model.isEmpty() ) {
                return null;
            }
            return new FastNumber( ((TemplateNumberModel)model).getAsNumber() );
        }
        if( model instanceof TemplateScalarModel ) {
            if( model.isEmpty() ) {
                return null;
            }
            long value;
            try {
                value = Long.parseLong( ((TemplateScalarModel)model).getAsString() );
            } catch( NumberFormatException e ) {
                throw new TemplateModelException( "Could not parse string as number", e );
            }
            return new FastNumber( value );
        }

        throw new TemplateModelException( "ToNumber must be given a string or number argument" );
    }

    /**
     * Is this model empty?
     *
     * @return <code>false</code> to indicate that this model is non-empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
