/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.regex.MatchResult;
import java.io.*;
import freemarker.template.TemplateTransformModel2;
import freemarker.template.TemplateModelException;

/**
 * <p>A FM-Classic transformer that performs syntax colouring of FM-Classic
 * source files, and outputs the results in HTML. The HTML is HTML 3.2 compliant,
 * though later transitional DTDs should have no problems with this code.</p>
 *
 * <p>The transformation uses the Jakarta ORO regular expression library,
 * available from the <a href="http://jakarta.apache.org/oro/"
 * target="_top">Jakarta Apache</a> web site.</p>
 *
 * @author  Nicholas Cull
 * @version $Id: FreeMarkerToHtml.java 1065 2004-11-06 05:14:07Z run2000 $
 */
public final class FreeMarkerToHtml implements TemplateTransformModel2, Serializable {
    private transient Perl5Util cMatcher;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -5357258851346554174L;

    /** Creates new FreeMarkerToHtml */
    public FreeMarkerToHtml() {
        cMatcher = new Perl5Util();
    }

    /**
     * Perform all FM-Classic template transformations.
     */
    private void doFreeMarkerTransformation(String textBuffer, Writer cOut)
            throws IOException {

        textBuffer = cMatcher.substitute( "s/&/&amp;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/</&lt;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/>/&gt;/g", textBuffer );

        cOut.write( "<pre>" );
        doBlockComment( textBuffer, cOut );
        cOut.write( "</pre>" );
    }

    /**
     * Format FM-Classic block comments.
     */
    private void doBlockComment(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/&lt;comment\\b(?:.*?)&lt;\\/comment\\s*&gt;/s",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doFreeMarkerTags( aInput.substring( 0, cResult.beginOffset(0) ), cOut );
            cOut.write( "<font color=\"#008000\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerTags( aInput, cOut );
    }

    /**
     * Format FM-Classic start/standalone tags.
     */
    private void doFreeMarkerTags(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/(&lt;(?:list|if|else|elseif|switch|case|break|" +
                "default|assign|include|function|exit|call|compress|comment|noparse|" +
                "foreach|transform))\\b(.*?)(&gt;)/s",  aInput)) {
            cResult = cMatcher.getMatch();
            doFreeMarkerEndTags( aInput.substring( 0, cResult.beginOffset(0) ), cOut );
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.group( 1 ) );
            cOut.write( "</font>" );
            doDoubleQuotes( cResult.group( 2 ), cOut );
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.group( 3 ) );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerEndTags( aInput, cOut );
    }

    /**
     * Format FM-Classic end tags.
     */
    private void doFreeMarkerEndTags(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/(&lt;\\/(?:list|if|switch|" +
                "function|compress|comment|noparse|foreach|" +
                "transform))\\s*(&gt;)/s",  aInput)) {
            cResult = cMatcher.getMatch();
            doFreeMarkerVariableInstruction( aInput.substring( 0, cResult.beginOffset(0) ), cOut );
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.group( 0 ) );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerVariableInstruction( aInput, cOut );
    }

    /**
     * Format FM-Classic variable instructions.
     */
    private void doFreeMarkerVariableInstruction(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/(\\$\\{)(.*?)(\\})/s",  aInput)) {
            cResult = cMatcher.getMatch();
            cOut.write( aInput.substring( 0, cResult.beginOffset(0) ));
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.group( 1 ) );
            cOut.write( "</font>" );
            doDoubleQuotes( cResult.group( 2 ), cOut );
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.group( 3 ) );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        cOut.write( aInput );
    }

    /**
     * Format FM-Classic double-quoted strings.
     */
    private void doDoubleQuotes(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/\"(?:\\\\.|[^\"])*\"/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doDigits( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#808080\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDigits( aInput, cOut );
    }

    /**
     * Format FM-Classic number literals.
     */
    private void doDigits(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/\\b(?:\\d+|#true|#false)\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doOperator( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#ff0000\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOperator( aInput, cOut );
    }

    /**
     * Format FM-Classic operators.
     */
    private void doOperator(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/(?:[\\+\\-\\*\\/\\%\\[\\]\\{\\}\\=\\.\\(\\)\\;\\!]|&amp;)+/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doSpecialWords( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<b>" );
            cOut.write( aMatch );
            cOut.write( "</b>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doSpecialWords( aInput, cOut );
    }

    /**
     * Format FM-Classic special words, used in "foreach" and "list"
     * instructions.
     */
    private void doSpecialWords(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/\\b(?:as|in|eq|ne|gt|ge|lt|le|type)\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doOtherWords( aInput.substring( 0, cResult.beginOffset(0) ), cOut );
            cOut.write( "<font color=\"#ff8000\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOtherWords( aInput, cOut );
    }

    /**
     * Format FM-Classic variables.
     */
    private void doOtherWords(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/\\w+#?/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            cOut.write( aInput.substring( 0, cResult.beginOffset(0) ));
            cOut.write( "<font color=\"#804040\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        cOut.write( aInput );
    }

    /**
     * Performs a transformation/filter on FM-Classic output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, Writer output)
            throws IOException, TemplateModelException {
        StringBuffer cText = new StringBuffer();
        char cBuffer[] = new char[ 1024 ];
        int nSize;

        nSize = source.read( cBuffer );
        while( nSize >= 0 ) {
            cText.append( cBuffer, 0, nSize );
            nSize = source.read( cBuffer );
        }
        doFreeMarkerTransformation( cText.toString(), output );
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, to indicate this object is not empty
     */
    public boolean isEmpty() {
        return false;
    }

    /**
     * For serialization, read this object normally, then reconstruct
     * the Perl 5 utility class.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        cMatcher = new Perl5Util();
    }
}
