/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.Serializable;
import java.io.ObjectStreamException;

import freemarker.template.*;

/**
 * <p>Determines whether a given model is an instance of a
 * <code>TemplateListModel</code> or <code>TemplateListModel2</code>.</p>
 *
 * <p><b>Note:</b><br />
 * <code>IsList</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: IsList.java 1025 2004-10-20 11:04:40Z run2000 $
 * @since 1.8
 */
public final class IsList implements TemplateMethodModel2, Serializable {

    private static final IsList instance = new IsList();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 2414796660481301168L;

    /** Private constructor, indicating that this is a singleton. */
    private IsList() {
    }

    /**
     * Return a singleton instance of this class.
     *
     * @return a new IsList object
     */
    public static IsList getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>true</code> if this object is empty, otherwise <code>false</code>
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * {@link TemplateModel} objects.
     *
     * @param arguments a <code>List</code> of <code>TemplateModel</code> objects
     * containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        if( arguments.size() != 1 ) {
            throw new TemplateModelException(
                    "IsList method needs exactly one argument" );
        }
        TemplateModel model = arguments.get(0);
        boolean match = ( model instanceof TemplateListModel ) ||
                        ( model instanceof TemplateListModel2 );
        return FastBoolean.getInstance( match );
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
