/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.regex.MatchResult;
import java.io.*;
import freemarker.template.TemplateTransformModel2;
import freemarker.template.TemplateModelException;

/**
 * <p>A FM-Classic transformer that performs syntax colouring of Java source
 * files, and outputs the results in HTML. The HTML is HTML 3.2 compliant,
 * though later transitional DTDs should have no problems with this code.</p>
 *
 * <p>The transformation uses the Jakarta ORO regular expression library,
 * available from the <a href="http://jakarta.apache.org/oro/"
 * target="_top">Jakarta Apache</a> web site.</p>
 *
 * @author  Nicholas Cull
 * @version $Id: JavaToHtml.java 1065 2004-11-06 05:14:07Z run2000 $
 */
public final class JavaToHtml implements TemplateTransformModel2, Serializable {
    private transient Perl5Util cMatcher;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 724356865491137221L;

    /** Creates new JavaToHtml */
    public JavaToHtml() {
        cMatcher = new Perl5Util();
    }

    /**
     * Perform all Java language transformations.
     */
    private void doJavaTransformation( String textBuffer, Writer cOut ) throws IOException {

        textBuffer = cMatcher.substitute( "s/&/&amp;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/</&lt;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/>/&gt;/g", textBuffer );

        cOut.write( "<pre>" );
        doBlockComment( textBuffer, cOut );
        cOut.write( "</pre>" );
    }

    /**
     * Format Java block comments.
     */
    private void doBlockComment(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\/\\*(?:.*?)\\*\\//s",  aInput)) {
            cResult = cMatcher.getMatch();
            doDoubleQuotes( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#008000\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDoubleQuotes( aInput, cOut );
    }

    /**
     * Format double quoted strings.
     */
    private void doDoubleQuotes(String aInput, Writer cOut) throws IOException {
        String aMatch;
        MatchResult cResult;

        while(cMatcher.match( "m/\"(?:\\\\.|[^\"])*\"/m",  aInput)) {
            cResult = cMatcher.getMatch();
            aMatch = cResult.toString();
            doSingleQuotes( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#808080\">" );
            cOut.write( aMatch );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doSingleQuotes( aInput, cOut );
    }

    /**
     * Format Java single quotes (ie. character data).
     */
    private void doSingleQuotes(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/'(?:\\\\.|[^\'])*'/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doLineComment( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#808080\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doLineComment( aInput, cOut );
    }

    /**
     * Format Java line comments (// syntax).
     */
    private void doLineComment(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\/\\/.*?$/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doPrimitiveTypes( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#008000\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doPrimitiveTypes( aInput, cOut );
    }

    /**
     * Format Java primitive types.
     */
    private void doPrimitiveTypes(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\b(?:boolean|byte|char|double|float|int|" +
                "long|short|void)\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doKeywords( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#ff8000\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doKeywords( aInput, cOut );
    }

    /**
     * Format Java keywords.
     */
    private void doKeywords(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\b(?:abstract|break|case|catch|" +
                "class|continue|default|do|else|extends|" +
                "false|final|finally|for|" +
                "if|implements|import|instanceof|interface|length|" +
                "native|new|null|package|private|protected|public|" +
                "return|static|super|switch|synchronized|" +
                "this|threadsafe|throw|throws|transient|true|try|" +
                "volatile|while" +
                ")\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doJavaPackages( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#0000ff\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doJavaPackages( aInput, cOut );
    }

    /**
     * Format three of the main Java classes: Java.lang, Java.io, and Java.util.
     */
    private void doJavaPackages(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\b(?:AbstractCollection|AbstractList|AbstractMap|" +
                "AbstractSequentialList|AbstractSet|ArrayList|Arrays|" +
                "BitSet|Boolean|BufferedInputStream|" +
                "BufferedOutputStream|BufferedReader|BufferedWriter|Byte|" +
                "ByteArrayInputStream|ByteArrayOutputStream|Calendar|Character|" +
                "Class|CharArrayReader|CharArrayWriter|ClassLoader|Cloneable|" +
                "Collection|Collections|Comparable|Comparator|" +
                "Compiler|DataInput|DataInputStream|DataOutput|DataOutputStream|" +
                "Date|Dictionary|Double|Enumeration|EventListener|EventObject|" +
                "Externalizable|File|FileDescriptor|FileFilter|FilenameFilter|" +
                "FileInputStream|FileOutputStream|FilePermission|FileReader|" +
                "FileWriter|FilterInputStream|FilterOutputStream|FilterReader|" +
                "FilterWriter|Float|GregorianCalendar|HashMap|HashSet|Hashtable|" +
                "InheritableThreadLocal|InputStream|InputStreamReader|Integer|" +
                "Iterator|LineNumberInputStream|LineNumberReader|LinkedList|" +
                "List|ListIterator|ListResourceBundle|Locale|Long|Map|Math|" +
                "Number|Object|ObjectInput|ObjectInputStream|ObjectInputValidation|" +
                "ObjectOutput|ObjectOutputStream|ObjectStreamClass|" +
                "ObjectStreamConstants|ObjectStreamField|Observable|Observer|" +
                "OutputStream|OutputStreamWriter|Package|PipedInputStream|" +
                "PipedOutputStream|PipedReader|PipedWriter|PrintStream|" +
                "PrintWriter|Process|Properties|PropertyPermission|" +
                "PropertyResourceBundle|PushbackInputStream|PushbackReader|" +
                "Random|RandomAccessFile|Reader|ResourceBundle|Runnable|" +
                "Runtime|RuntimePermission|SecurityManager|SequenceInputStream|" +
                "Serializable|SerializablePermission|Set|Short|SimpleTimeZone|" +
                "SortedMap|SortedSet|Stack|StreamTokenizer|StrictMath|String|" +
                "StringBuffer|StringBufferInputStream|StringReader|" +
                "StringTokenizer|StringWriter|System|Thread|ThreadGroup|" +
                "ThreadLocal|Timer|TimerTask|TimeZone|TreeMap|TreeSet|" +
                "Vector|Void|WeakHashMap|Writer" +
                ")\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doDigits( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#804040\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDigits( aInput, cOut );
    }

    /**
     * Format digits in Java source code.
     */
    private void doDigits(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\b\\d+\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doOperator( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<font color=\"#ff0000\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOperator( aInput, cOut );
    }

    /**
     * Format Java operators.
     */
    private void doOperator(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/(?:\\w+\\s*\\(|[\\+\\*\\^\\$\\-\\{\\}\\[\\]" +
                "\\=\\.\\(\\)\\,\\:\\/\\;]|&(lt|gt|amp);" +
                ")+/m",  aInput)) {
            cResult = cMatcher.getMatch();
            doExceptions( aInput.substring( 0, cResult.beginOffset(0) ), cOut);
            cOut.write( "<b>" );
            cOut.write( cResult.toString() );
            cOut.write( "</b>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doExceptions( aInput, cOut );
    }

    /**
     * Format Java exception classes.
     */
    private void doExceptions(String aInput, Writer cOut) throws IOException {
        MatchResult cResult;

        while(cMatcher.match( "m/\\w*(Error|Exception|Throwable)\\b/m",  aInput)) {
            cResult = cMatcher.getMatch();
            cOut.write( aInput.substring( 0, cResult.beginOffset(0) ));
            cOut.write( "<font color=\"#ff0000\">" );
            cOut.write( cResult.toString() );
            cOut.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        cOut.write( aInput );
    }

    /**
     * Performs a transformation/filter on FM-Classic output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, Writer output)
            throws IOException, TemplateModelException {
        StringBuffer cText = new StringBuffer();
        char cBuffer[] = new char[ 1024 ];
        int nSize;

        nSize = source.read( cBuffer );
        while( nSize >= 0 ) {
            cText.append( cBuffer, 0, nSize );
            nSize = source.read( cBuffer );
        }
        doJavaTransformation( cText.toString(), output );
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, to indicate this object is not empty
     */
    public boolean isEmpty() {
        return false;
    }

    /**
     * For serialization, read this object normally, then reconstruct
     * the Perl 5 utility class.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        cMatcher = new Perl5Util();
    }
}
