/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.util.Iterator;
import java.io.Serializable;

import freemarker.template.*;
import freemarker.template.expression.ExpressionUtils;

/**
 * Replicates the standard Perl join function. Partners with the Split
 * method model. Any sublists are joined along with the main list in a
 * depth-first traversal order.
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "join", freemarker.ext.misc.Join.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following creates a directory path:
 * &lt;assign path = [ "var", "log", "apache" ]&gt;
 * &lt;assign output = join( "/", path )&gt;
 * /${output}
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: Join.java 1185 2005-10-13 12:24:14Z run2000 $
 * @since 1.9
 */
public class Join implements TemplateMethodModel2, Serializable {
    private static final Join instance = new Join();

    /** Serialization UUID for this class. */
    private static final long serialVersionUID = 5344842626008513409L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private Join() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>Join</code> class
     */
    public static Join getInstance() {
        return instance;
    }

    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {

        if( arguments.size() < 2 ) {
            return null;
        }

        TemplateModel separatorModel = arguments.get( 0 );
        String separator;

        try {
            separator = ExpressionUtils.getAsStringOrEmpty( separatorModel );
        } catch( TemplateException e ) {
            throw new TemplateModelException( e.getMessage(), e );
        }

        List<TemplateModel> items = arguments.subList( 1, arguments.size() );
        Iterator<TemplateModel> iItems = items.iterator();
        StringBuffer result = new StringBuffer();

        while( iItems.hasNext() ) {
            TemplateModel item = iItems.next();
            if( item instanceof TemplateListModel ) {
                result.append( appendList( separator, (TemplateListModel) item ));
            } else if( item instanceof TemplateListModel2 ) {
                result.append( appendList( separator, (TemplateListModel2) item ));
            } else {
                try {
                    result.append( ExpressionUtils.getAsStringOrEmpty( item ));
                } catch( TemplateException e ) {
                    throw new TemplateModelException( e.getMessage(), e );
                }
            }
            if(( iItems.hasNext() ) && ( separator.length() > 0 )) {
                result.append( separator );
            }
        }

        return new FastScalar( result.toString() );
    }

    /**
     * Iterate over TemplateListModel2 items and join the results together.
     * This may in turn involve recursing over child list models.
     *
     * @param separator the separator String to place between each item
     * @param items the items to be joined, possibly containing sublists
     * @return a String containing the joined items
     * @throws TemplateModelException there was a problem with one of the
     * models
     */
    private String appendList( String separator, TemplateListModel2 items )
            throws TemplateModelException {
        TemplateIteratorModel iItems = items.templateIterator();
        StringBuffer result = new StringBuffer();

        while( iItems.hasNext() ) {
            TemplateModel item = iItems.next();
            if( item instanceof TemplateListModel ) {
                result.append( appendList( separator, (TemplateListModel) item ));
            } else if( item instanceof TemplateListModel2 ) {
                result.append( appendList( separator, (TemplateListModel2) item ));
            } else {
                try {
                    result.append( ExpressionUtils.getAsStringOrEmpty( item ));
                } catch( TemplateException e ) {
                    throw new TemplateModelException( e.getMessage(), e );
                }
            }
            if(( iItems.hasNext() ) && ( separator.length() > 0 )) {
                result.append( separator );
            }
        }
        return result.toString();
    }

    /**
     * Iterate over TemplateListModel items and join the results together.
     * This may in turn involve recursing over child list models.
     *
     * @param separator the separator String to place between each item
     * @param items the items to be joined, possibly containing sublists
     * @return a String containing the joined items
     * @throws TemplateModelException there was a problem with one of the
     * models
     */
    private String appendList( String separator, TemplateListModel items )
            throws TemplateModelException {

        StringBuffer result = new StringBuffer();

        if( ! items.isRewound() ) {
            items.rewind();
        }

        while( items.hasNext() ) {
            TemplateModel item = items.next();
            if( item instanceof TemplateListModel ) {
                result.append( appendList( separator, (TemplateListModel) item ));
            } else if( item instanceof TemplateListModel2 ) {
                result.append( appendList( separator, (TemplateListModel2) item ));
            } else {
                try {
                    result.append( ExpressionUtils.getAsStringOrEmpty( item ));
                } catch( TemplateException e ) {
                    throw new TemplateModelException( e.getMessage(), e );
                }
            }
            if(( items.hasNext() ) && ( separator.length() > 0 )) {
                result.append( separator );
            }
        }

        return result.toString();
    }
}
