/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import org.apache.log4j.Priority;
import org.apache.log4j.Category;
import freemarker.template.*;

/**
 * <p>Adapts FM-Classic's exception events to Log4j's logging methods. This
 * code is known to work with Log4j version 1.1.3, but should be compatible
 * with earlier versions as well.</p>
 *
 * <p>Log4j can be found at the <a href="http://jakarta.apache.org/log4j/"
 * target="_top">Jakarta Apache</a> web site.</p>
 *
 * <p><b>Usage example:</b></p>
 * <pre>
 * // Configure Log4j
 * Category rootCategory = Category.getRoot();
 * PatternLayout patternLayout = new PatternLayout(PatternLayout.TTCC_CONVERSION_PATTERN);
 * ConsoleAppender consoleAppender = new ConsoleAppender(patternLayout, "System.err");
 * rootCategory.addAppender(consoleAppender);
 *
 * // Set up the template
 * Template template = new Template(filename);
 *
 * // Set up the model
 * TemplateModelRoot modelRoot = (...)
 *
 * // Set up the listener
 * SimpleEventAdapter adapter = new SimpleEventAdapter( new Log4jExceptionListener() );
 *
 * // Call the template
 * template.process( modelRoot, writer, adapter );
 * </pre>
 *
 * @author  Nicholas Cull
 * @version $Id: Log4jExceptionListener.java 1065 2004-11-06 05:14:07Z run2000 $
 */
public final class Log4jExceptionListener implements TemplateExceptionListener {

    /** Creates new Log4jExceptionListener */
    public Log4jExceptionListener() {
    }

    /**
     * Passes exceptions raised by FreeMarker onto Log4j. The categories
     * supplied to Log4j match the class and method name that caused the
     * event to be fired.
     */
    public void exceptionThrown(TemplateExceptionEvent e) {
        Category cat = Category.getInstance( e.getSourceName() );
        StringBuffer message = new StringBuffer();
        Exception ex = e.getException();

        message.append( ex.getMessage() );
        if( ex instanceof TemplateException ) {
            Exception cause = ((TemplateException)ex).getCauseException();
            if( cause != null ) {
                message.append( ": " );
                message.append( cause.getMessage() );
            }
        }
        Priority priority = null;

        switch( e.getSeverity() ) {
            case ERROR:
                priority = Priority.ERROR;
                break;

            case WARNING:
                priority = Priority.WARN;
                break;

            case DEPRECATION:
                priority = Priority.INFO;
                break;
        }

        cat.log( priority, message.toString() );
    }

}
