/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.TemplateHashModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;

import java.io.Serializable;
import java.io.ObjectStreamException;

/**
 * <p>A simple hash model that provides access to all the non-deprecated method
 * models in this package.</p>
 *
 * <p>The keys are as follows:</p>
 * <ul>
 * <li>asNumber -- The {@link AsNumber} method</li>
 * <li>asString -- the {@link AsString} method</li>
 * <li>execute -- the {@link Execute} method</li>
 * <li>hashKeys -- the {@link HashKeys} method</li>
 * <li>hashValues -- the {@link HashValues} method</li>
 * <li>isEmpty -- the {@link IsEmpty} method</li>
 * <li>isHash -- the {@link IsHash} method</li>
 * <li>isIndexed -- the {@link IsIndexed} method</li>
 * <li>isList -- the {@link IsList} method</li>
 * <li>isMethod -- the {@link IsMethod} method</li>
 * <li>isNumber -- the {@link IsNumber} method</li>
 * <li>isScalar -- the {@link IsScalar} method</li>
 * <li>isTransform -- the {@link IsTransform} method</li>
 * <li>join -- the {@link Join} method</li>
 * <li>size -- the {@link CollectionSize} method</li>
 * <li>translate -- the {@link Translate} method</li>
 * </ul>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "methods", freemarker.ext.misc.Methods.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * &lt;assign var1 = "5"&gt;
 * &lt;assign var2 = "3"&gt;
 * &lt;assign var3 = methods.asNumber( var1 ) + methods.asNumber( var2 )&gt;
 * &lt;p&gt;The sum of ${var1} and ${var2} is: ${var3}&lt;/p&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>Methods</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @version $Id: Methods.java 1093 2005-08-30 12:23:57Z run2000 $
 * @author Nicholas Cull
 * @since 1.8.1
 */
public class Methods implements TemplateHashModel, Serializable {
    private static final Methods instance = new Methods();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -9010210904654477131L;

    /** Private constructor, to indicate this is a singleton class. */
    private Methods() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>Methods</code> class
     */
    public static Methods getInstance() {
        return instance;
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code> to indicate that this model is non-empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Gets a {@link TemplateModel} from the hash.
     *
     * @param key the name by which the <code>TemplateModel</code>
     * is identified in the template.
     * @return the <code>TemplateModel</code> referred to by the key,
     * or <code>null</code> if not found.
     */
    public TemplateModel get( String key ) throws TemplateModelException {
        if( key == null ) {
            return null;
        } else if( "asNumber".equals(key)) {
            return AsNumber.getInstance();
        } else if( "asString".equals(key)) {
            return AsString.getInstance();
        } else if( "size".equals(key)) {
            return CollectionSize.getInstance();
        } else if( "execute".equals(key)) {
            return Execute.getInstance();
        } else if( "hashKeys".equals(key)) {
            return HashKeys.getInstance();
        } else if( "hashValues".equals(key)) {
            return HashValues.getInstance();
        } else if( "isEmpty".equals(key)) {
            return IsEmpty.getInstance();
        } else if( "isHash".equals(key)) {
            return IsHash.getInstance();
        } else if( "isIndexed".equals(key)) {
            return IsIndexed.getInstance();
        } else if( "isList".equals(key)) {
            return IsList.getInstance();
        } else if( "isMethod".equals(key)) {
            return IsMethod.getInstance();
        } else if( "isNumber".equals(key)) {
            return IsNumber.getInstance();
        } else if( "isScalar".equals(key)) {
            return IsScalar.getInstance();
        } else if( "isTransform".equals(key)) {
            return IsTransform.getInstance();
        } else if( "join".equals(key)) {
            return Join.getInstance();
        }

        // Libraries requiring the Jakarta ORO library
        if( "match".equals(key)) {
            return new Match();
        } else if( "split".equals(key)) {
            return new Split();
        } else if( "substitute".equals(key)) {
            return new Substitute();
        }

        // Libraries requiring the JTR library
        if( "translate".equals(key)) {
            return Translate.getInstance();
        }

        return null;
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
