/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.misc;

import java.io.*;

import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateModelException;

/**
 * <p>Transformer that supports FreeMarker legacy behaviour: all newlines appearing
 * within the transformed area will be transformed into the platform's default
 * newline. Unlike the old behaviour, however, newlines generated by the data
 * model are also converted. Legacy behaviour was to leave newlines in the
 * data model unaltered.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "normalizeNewlines", freemarker.ext.misc.NormalizeNewlines.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * &lt;transform normalizeNewlines&gt;
 *   &lt;html&gt;
 *   &lt;head&gt;
 *   ...
 *   &lt;p&gt;This template has all newlines normalized to the current platform's
 *   default.&lt;/p&gt;
 *   ...
 *   &lt;/body&gt;
 *   &lt;/html&gt;
 * &lt;/transform&gt;
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>NormalizeNewlines</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @version $Id: NormalizeNewlines.java 1179 2005-10-10 13:42:57Z run2000 $
 */
public final class NormalizeNewlines implements TemplateTransformModel,
        Serializable {

    private static final NormalizeNewlines instance = new NormalizeNewlines();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -1064146228486518529L;

    /** Private constructor, to indicate this is a singleton class. */
    private NormalizeNewlines() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>NormalizeNewlines</code> class
     */
    public static NormalizeNewlines getInstance() {
        return instance;
    }

    /**
     * Performs newline normalization on FM-Classic output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, PrintWriter output) throws TemplateModelException {
        BufferedReader br = new BufferedReader(source);
        String line;

        try {
            if((line = br.readLine()) != null) {
                if( line.length() > 0 ) {
                    output.println( line );
                }
            }
            while ((line = br.readLine()) != null) {
                output.println( line );
            }
        } catch (IOException e) {
            try {
                br.close();
            } catch (IOException e1) { }
            throw new TemplateModelException( e.getMessage() );
        }

        try {
            br.close();
        } catch( IOException e ) {
            throw new TemplateModelException( e.getMessage() );
        }
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code>, to indicate this object is not empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
