/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.util.Arrays;
import java.io.*;

import freemarker.template.TemplateMethodModel2;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateTransformModel2;

/**
 * <p>A class for chaining multiple transformations into a single model without
 * having to nest &lt;transform&gt; tags. A slightly more elegant way to
 * handle transformation pipelines.</p>
 *
 * <p>Transformations are called from leftmost argument to rightmost in order.
 * This differs from nesting &lt;transform&gt; tags where the innermost is
 * called first, and the outermost is called last.</p>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "pipeline", freemarker.ext.misc.RemoveBlankLines.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is transformed with a pipeline:
 * &lt;transform pipeline( transform1, transform2, transform3, transform4 )&gt;
 *   This output is transformed by all the transformations in the pipeline.
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p>In the example, the transformatiosn are applied in the following order:</p>
 * <p> Source =&gt; Transform 1 =&gt; Transform 2 =&gt; Transform 3 =&gt;
 * Transform 4 = &gt; Output</p>
 *
 * <p><b>Note:</b><br />
 * <code>Pipeline</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: Pipeline.java 1102 2005-09-07 21:36:39Z run2000 $
 * @since 1.9
 */
public final class Pipeline implements TemplateMethodModel2, Serializable {

    private static final Pipeline instance = new Pipeline();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -1492993253658875861L;

    /**
     * Private constructor indicating the Pipeline is a singleton.
     * Note however, that the transformation pipeline class itself is
     * instantiated for each new call to the {@link #exec} method.
     */
    private Pipeline() { }

    /**
     * Return a new instance of this class. No state is held in the method
     * model itself. Rather, the method model creates a new instance
     * of the transformation class itself whenever it is called. The
     * transformation class holds all state information about the pipeline.
     *
     * @return a new instance of the Pipeline method model
     */
    public static Pipeline getInstance() {
        return instance;
    }

    /**
     * Is the current method model empty.
     *
     * @return <code>false</code>, indicating that this method model is not
     * empty
     */
    public boolean isEmpty() {
        return false;
    }

    /**
     * Construct a new transform pipeline with the given transform models.
     * The models will be called in leftmost to rightmost order.
     *
     * @param arguments a List of TemplateTransformModel or
     * TemplateTransformModel2 models to be chained into a single pipeline
     * @return a new TemplateTransformModel2 that represents all the
     * transform models in a single pipeline
     * @throws TemplateModelException
     */
    public TemplateModel exec( List<TemplateModel> arguments ) throws TemplateModelException {
        // Need to determine up-front if all list arguments are
        // TemplateTransformModel or TemplateTransformModel2 implementations
        int size = arguments.size();
        TemplateModel model;

        for( int i = 0; i < size; i++ ) {
            model = (TemplateModel) arguments.get( i );
            if((model != null) && !( model instanceof TemplateTransformModel ) && !( model instanceof TemplateTransformModel2 )) {
                throw new TemplateModelException( "Argument " + i +
                        " is not a transform model, it is a " + model.getClass().getName() );
            }
        }

        return new TransformPipeline( arguments );
    }

    /**
     * A pipeline of transform models that is itself a transform model.
     * Transformations occur in leftmost to rightmost order. Null or
     * empty models are considered equivalent to the null transformation,
     * where data is simply copied from input to output without alteration.
     * Similarly, a pipeline consisting of no transformations is also handled
     * as a null tranformation.
     */
    private static class TransformPipeline implements TemplateTransformModel2, Serializable {
        /**
         * @serial the list of transform models to be pipelined. In practise,
         * serialization depends on all the contained models also being
         * serializable.
         */
        private TemplateModel[] transforms;

        /** Class UUID for serialization. */
        private static final long serialVersionUID = -3179369201409975041L;

        /**
         * Create a new transform pipeline with the given transform models.
         *
         * @param transforms an ordered list of TemplateTransformModel and
         * TemplateTransformModel2 objects to be pipelined
         */
        public TransformPipeline( List<TemplateModel> transforms ) {
            this.transforms = new TemplateModel[ transforms.size() ];
            transforms.toArray( this.transforms );
        }

        /**
         * Is there anything in this pipeline.
         *
         * @return <code>true</code> if the pipeline is empty, otherwise
         * <code>false</code>
         * @throws TemplateModelException
         */
        public boolean isEmpty() throws TemplateModelException {
            return ( transforms == null ) || ( transforms.length == 0 );
        }

        /**
         * Perform the requested transformations on the given data source in
         * the given order.
         *
         * @param source the input data to be transformed
         * @param target the data after being fed through all transformations
         * @throws IOException
         * @throws TemplateModelException
         */
        public void transform( Reader source, Writer target )
                throws IOException, TemplateModelException {

            // This shouldn't be necessary, but deal with empty transformations
            // just in case.
            if(( transforms == null ) || ( transforms.length == 0 )) {
                pipelineStage( source, target, null);
                return;
            }

            // Two special cases to be considered:
            // * the beginning, where the input is the source reader
            // * the end, where the target is the output writer
            String intermediate = "";

            for( int i = 0; i < transforms.length; i++ ) {
                Reader input;

                if( i == 0 ) {
                    input = source;
                } else {
                    input = new StringReader( intermediate );
                }
                if( i + 1 == transforms.length ) {
                    pipelineStage( input, target, transforms[ i ]);
                } else {
                    StringWriter output = new StringWriter();
                    pipelineStage( input, output, transforms[ i ]);
                    intermediate = output.toString();
                }
            }
        }

        /**
         * Perform a single pipeline stage from the given source to the given
         * target using the supplied transformation. If we're given a
         * <code>null</code> or empty transform model, perform a null
         * transformation on the source. Otherwise, determine the type of model,
         * and call with the appropriate arguments.
         *
         * @param source the source data to be transformed
         * @param target the target where transformed data will be written
         * @param transform the transformation to be applied, possibly
         * <code>null</code> or empty
         * @throws IOException an input/output error occured during the
         * transformation
         * @throws TemplateModelException the transformation threw an exception
         */
        private void pipelineStage( Reader source, Writer target,
                TemplateModel transform ) throws IOException, TemplateModelException {

            // Deal with null transformations
            if(( transform == null ) || ( transform.isEmpty() )) {
                int bufferLen = 1024;
                char[] buffer = new char[ bufferLen ];
                int size = 0;
                size = source.read( buffer );
                while( size > 0 ) {
                    target.write( buffer, 0, size );
                }
            }

            if( transform instanceof TemplateTransformModel ) {
                TemplateTransformModel pipeline = (TemplateTransformModel) transform;
                pipeline.transform( source, new PrintWriter( target ));
            } else if( transform instanceof TemplateTransformModel2 ) {
                TemplateTransformModel2 pipeline = (TemplateTransformModel2) transform;
                pipeline.transform( source, target );
            }
        }

        /**
         * Returns a string representation of the object.
         *
         * @return a <code>String</code> representation of the object.
         */
        public String toString() {
            StringBuffer buffer = new StringBuffer();

            buffer.append( "Pipelined transforms: " );
            buffer.append( Arrays.asList( transforms ));
            return buffer.toString();
        }
    }
}
