/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.IOException;

import freemarker.template.TemplateMethodModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;
import freemarker.template.FastScalar;
import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.perl.MalformedPerl5PatternException;
import org.apache.oro.text.PatternCacheLRU;

/**
 * <p>A method model that implements the Perl 5 substitute function. This uses the
 * Jakarta ORO library to perform the substitute.</p>
 *
 * <p><b>Usage:</b><br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "substitute", new freemarker.ext.misc.Substitute() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is stripped of HTML tags:
 * &lt;assign text = "&lt;p&gt;This paragraph has &lt;em&gt;all&lt;/em&gt; tags removed.&lt;/p&gt;"&gt;
 * &lt;assign output = substitute( "s/&lt;[^&gt;]*&gt;//g", text )&gt;
 *
 * ${output}
 *
 * ...
 * </pre>
 *
 * @author Nicholas Cull
 * @version $Id: Substitute.java 1093 2005-08-30 12:23:57Z run2000 $
 * @since 1.8.2
 */
public final class Substitute implements TemplateMethodModel, Serializable {

    private transient PatternCacheLRU patternCache;

    /** Class UUID for serialization. */
    private static final long serialVersionUID = -652420927008297709L;

    /**
     * Construct a new substitution method.
     */
    public Substitute() {
        patternCache = new PatternCacheLRU();
    }

    /**
     * Is the object empty?
     *
     * @return <code>false</code> to indicate the method is not empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Executes a method call. Arguments are passed as a <code>List</code> of
     * <code>String</code> objects.
     *
     * @param arguments a <code>List</code> of <code>String</code> objects
     * containing the values of the arguments passed to the method.
     * @return the <code>TemplateModel</code> produced by the method, or <code>null</code>.
     */
    public TemplateModel exec( List<String> arguments ) throws TemplateModelException {

        if( arguments.size() != 2 ) {
            throw new TemplateModelException( "Substitute needs two arguments" );
        }

        Perl5Util perl5Util = new Perl5Util(patternCache);
        try {
            String result = perl5Util.substitute( arguments.get(0),
                                                  arguments.get(1) );
            return new FastScalar( result );
        } catch( MalformedPerl5PatternException e ) {
            throw new TemplateModelException(
                    "Perl 5 regular expression could not be compiled: "
                    + e.getMessage(), e );
        }
    }

    /**
     * For serialization, read this object normally, then reconstruct
     * the Perl 5 pattern cache utility class.
     *
     * @param stream the input stream to read serialized objects from
     */
    private void readObject(ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        stream.defaultReadObject();
        patternCache = new PatternCacheLRU();
    }
}
