/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import freemarker.template.TemplateHashModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModel;

import java.io.Serializable;
import java.io.ObjectStreamException;

/**
 * <p>A simple hash model that provides access to all the non-deprecated method
 * models in this package.</p>
 *
 * <p>The keys are as follows:</p>
 * <ul>
 * <li>chain -- the {@link Pipeline} transform</li>
 * <li>compress -- the {@link CompressWhitespace} transform</li>
 * <li>freemarkerToHtml -- the {@link FreeMarkerToHtml} transform</li>
 * <li>htmlEscape -- the {@link HtmlEscape} transform</li>
 * <li>javaToHtml -- the {@link JavaToHtml} transform</li>
 * <li>legacyCompress -- the {@link LegacyCompress} transform</li>
 * <li>normalizeNewlines -- the {@link NormalizeNewlines} transform</li>
 * <li>pipeline -- the {@link Pipeline} transform</li>
 * <li>removeBlankLines -- the {@link RemoveBlankLines} transform</li>
 * <li>substitute -- the {@link Perl5Substitute} transform</li>
 * <li>translate -- the {@link Perl5Translate} transform</li>
 * <li>xmlEscape -- the {@link XmlEscape} transform</li>
 * </ul>
 *
 * <p>Usage:<br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "transforms", freemarker.ext.misc.Transforms.getInstance() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is compressed:
 * &lt;transform transforms.compressWhitespace&gt;
 *   &lt;p&gt;This paragraph has all whitespace reduced to a single
 *   space character.&lt;/p&gt;
 * &lt;/transform&gt;
 *
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>Transforms</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @version $Id: Transforms.java 1093 2005-08-30 12:23:57Z run2000 $
 * @author Nicholas Cull
 * @since 1.8.1
 */
public class Transforms implements TemplateHashModel, Serializable {
    private static final Transforms instance = new Transforms();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 5997924300246398113L;

    /** Private constructor, to indicate this is a singleton class. */
    private Transforms() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>Transforms</code> class
     */
    public static Transforms getInstance() {
        return instance;
    }


    /**
     * Is the object empty?
     *
     * @return <code>false</code> to indicate that this model is non-empty
     */
    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    /**
     * Gets a {@link TemplateModel} from the hash.
     *
     * @param key the name by which the <code>TemplateModel</code>
     * is identified in the template.
     * @return the <code>TemplateModel</code> referred to by the key,
     * or <code>null</code> if not found.
     */
    public TemplateModel get( String key ) throws TemplateModelException {

        if( key == null ) {
            return null;
        } else if( "compress".equals(key)) {
            return CompressWhitespace.getInstance();
        } else if( "htmlEscape".equals(key)) {
            return HtmlEscape.getInstance();
        } else if( "legacyCompress".equals(key)) {
            return LegacyCompress.getInstance();
        } else if( "normalizeNewlines".equals(key)) {
            return NormalizeNewlines.getInstance();
        } else if( "removeBlankLines".equals(key)) {
            return RemoveBlankLines.getInstance();
        } else if( "xmlEscape".equals(key)) {
            return XmlEscape.getInstance();
        }

        // Libraries requiring the Jakarta ORO library
        if( "freemarkerToHtml".equals(key)) {
            return new FreeMarkerToHtml();
        } else if( "javaToHtml".equals(key)) {
            return new JavaToHtml();
        } else if( "substitute".equals(key)) {
            return Perl5Substitute.getInstance();
        }

        return null;
    }

    /**
     * For serialization purposes, always resolve a de-serialized object
     * back to the singleton instance. This ensures that any strict equality
     * tests remain valid after serialization.
     */
    private Object readResolve() throws ObjectStreamException {
        return instance;
    }
}
