/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2005 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.misc;

import java.util.List;
import java.io.Serializable;

import freemarker.template.*;

import org.jtr.transliterate.Perl5Parser;
import org.jtr.transliterate.CharacterReplacer;
import org.jtr.transliterate.CharacterParseException;

/**
 * <p>A method model that implements the Perl 5 translate function. This uses
 * the JTR library to perform the translation.</p>
 *
 * <p><b>Usage:</b><br />
 * From java:</p>
 * <pre>
 * TemplateModelRoot root = new SimpleHash();
 *
 * root.put( "translate", new freemarker.ext.misc.Translate() );
 *
 * ...
 * </pre>
 *
 * <p>From your FM-Classic template:</p>
 * <pre>
 * The following is stripped of all duplicate spaces:
 * &lt;assign text = "   This paragraph has  all  excess  spaces  removed."&gt;
 * &lt;assign output = translate( "tr/ //s", text )&gt;
 *
 * ${output}
 *
 * The result is ${output.result}, the number of matches was ${output.matches}.
 * ...
 * </pre>
 *
 * <p><b>Note:</b><br />
 * <code>Translate</code> is a singleton instance. Use the
 * {@link #getInstance()} method to retrieve instances of this model.</p>
 *
 * @author Nicholas Cull
 * @version $Id: Translate.java 1093 2005-08-30 12:23:57Z run2000 $
 * @since 1.9
 */
public class Translate implements TemplateMethodModel, Serializable {
    private static final Translate instance = new Translate();

    /** Class UUID for serialization. */
    private static final long serialVersionUID = 1929541287462943061L;

    /**
     * Private constructor, meaning this shouldn't be instantiated directly.
     * Use the {@link #getInstance} method to return an instance of this
     * class.
     */
    private Translate() {
    }

    /**
     * Retrieve a singleton instance of this class. Since there is no
     * state information held between calls, there is only a need for
     * one instance.
     *
     * @return an instance of this <code>Translate</code> class
     */
    public static Translate getInstance() {
        return instance;
    }

    public boolean isEmpty() throws TemplateModelException {
        return false;
    }

    public TemplateModel exec( List<String> arguments ) throws TemplateModelException {

        if( arguments.size() != 2 ) {
            throw new TemplateModelException( "Translate needs two arguments" );
        }

        CharacterReplacer replacer;
        try {
            replacer = Perl5Parser.makeReplacer( arguments.get( 0 ));
        } catch( CharacterParseException e ) {
            throw new TemplateModelException( "Could not parse translation: " + e.getMessage(), e );
        }

        String result = replacer.doReplacement( arguments.get( 1 ));
        int matches = replacer.getMatches();
        return new TranslateResult( result, matches );
    }

    /**
     * Contains the result of the translation. The result can be returned
     * directly as a scalar model, or accessed using the "result" key. The
     * number of matches can be accessed using the "matches" key.
     */
    private static class TranslateResult implements TemplateScalarModel, TemplateHashModel {
        private String result;
        private int matches;

        public TranslateResult( String result, int matches ) {
            this.result = result;
            this.matches = matches;
        }

        public boolean isEmpty() throws TemplateModelException {
            return result == null;
        }

        public String getAsString() throws TemplateModelException {
            return result;
        }

        public TemplateModel get( String key ) throws TemplateModelException {
            if( "result".equals( key )) {
                return new FastScalar( result );
            } else if( "matches".equals( key )) {
                return new FastNumber( matches );
            }
            return null;
        }
    }
}
