/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.ext.util;

import java.io.*;

/**
 * <p>Splits the given input stream or filename(s) into one or more output
 * files. The splitter looks for lines containing the text:</p>
 * <pre>
 * &lt;!-- Filename: newfilename.blah --&gt;
 * </pre>
 * <p>When such a line is encountered, a new file output stream is created,
 * and the input is written to the named file until either the end of the
 * input stream, or another such tag is encountered. The filename can be
 * surrounded by optional double-quotes if leading or trailing whitespace
 * is required in the filename.</p>
 *
 * @author Nicholas Cull
 * @since 1.8.2
 * @version $Id: Splitter.java 1012 2004-10-18 05:47:34Z run2000 $
 */
public class Splitter {

    /** Private constructor, indicating this is a static class. */
    private Splitter() {
    }

    /**
     * Read an input stream and split it into one or more output files
     * based on comment tokens in the input.
     *
     * @param stream the input stream to be split into separate files
     * @throws IOException there was a problem with either input or output
     * streams
     */
    public static void splitInputStream( InputStream stream ) throws IOException {
        BufferedReader reader = new BufferedReader( new InputStreamReader( stream ));
        String line = reader.readLine();
        BufferedWriter writer = null;
        String currentFile;

        while( line != null ) {
            String temp = line.trim();
            if( temp.startsWith( "<!-- Filename: " ) && temp.endsWith( " -->" )) {
                currentFile = temp.substring( 15, temp.length() - 4 ).trim();
                if(( currentFile.length() > 2 ) && ( currentFile.charAt( 0 ) == '"' )
                        && ( currentFile.charAt( currentFile.length() - 1 ) == '"' )) {
                    currentFile = currentFile.substring( 1, currentFile.length() - 1 );
                }
                if( writer != null ) {
                    writer.close();
                }
                writer = new BufferedWriter( new FileWriter( currentFile ));
            } else if( writer != null ) {
                writer.write( line );
                writer.newLine();
            }
            line = reader.readLine();
        }

        reader.close();
        if( writer != null ) {
            writer.close();
        }
    }

    /**
     * Split the given file or files into one or more output files. If no
     * argument is specified, stream from stdin.
     *
     * @param args the input filename, or none to specify standard input
     */
    public static void main( String args[] ) throws IOException {
        if( args.length == 0 ) {
            splitInputStream( System.in );
        } else {
            for( String filename : args ) {
                File inputFile = new File( filename );
                if( ! inputFile.exists() ) {
                    System.err.println( "Cannot find file " + filename + ", ignoring." );
                } else {
                    FileInputStream stream = new FileInputStream( inputFile );
                    splitInputStream( stream );
                    stream.close();
                }
            }
        }
    }
}
