/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.ext.util;

import java.io.Writer;
import java.io.IOException;

/**
 * Implements a simple, non-synchronized character buffer as a <code>Writer</code>.
 * Since FM-Classic's write methods are all synchronous, we can get
 * away with unsynchronized calls. Sychronization of the underlying
 * <code>Writer</code> is left to the wrapped <code>Writer</code> itself.
 *
 * @author Nicholas Cull
 * @version $Id: TemplateBufferedWriter.java 1003 2004-10-16 14:54:58Z run2000 $
 * @since 1.8
 */
public final class TemplateBufferedWriter extends Writer {
    /** The underlying writer that we're writing to. */
    private Writer parentWriter;
    /** The character buffer for holding buffered output. */
    private char[] buffer;
    /** The offset for the next character in the buffer. */
    private int offset = 0;

    /** The default buffer size. */
    private static final int DEFAULT_BUFFER_SIZE = 2048;

    /**
     * Constructor that takes a <code>Writer</code> as a destination for
     * the buffered output.
     *
     * @param parentWriter the Writer that the buffered content should be
     * written to
     */
    public TemplateBufferedWriter(Writer parentWriter) {
        super(parentWriter);
        this.parentWriter = parentWriter;
        buffer = new char[ DEFAULT_BUFFER_SIZE ];
    }

    /**
     * Constructor that takes a <code>Writer</code> as a destination for
     * the buffered output, along with a size of the buffer to be used.
     *
     * @param parentWriter the Writer to which the buffered content should be
     * written
     * @param bufferSize the size of the buffer
     */
    public TemplateBufferedWriter(Writer parentWriter, int bufferSize) {
        super(parentWriter);
        this.parentWriter = parentWriter;
        buffer = new char[ bufferSize ];
    }

    /**
     * Write a portion of an array of characters.
     *
     * @param  cbuf  Array of characters
     * @param  off   Offset from which to start writing characters
     * @param  len   Number of characters to write
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void write(char cbuf[], int off, int len) throws IOException {
        int available = buffer.length - offset;
        if( len < available ) {
            System.arraycopy( cbuf, off, buffer, offset, len );
            offset += len;
            return;
        }
        System.arraycopy( cbuf, off, buffer, offset, available );
        parentWriter.write( buffer, 0, buffer.length );
        if( len - available >= buffer.length ) {
            parentWriter.write( cbuf, off + available, len - available );
            offset = 0;
        } else {
            offset = len - available;
            if( offset > 0 ) {
                System.arraycopy( cbuf, off + available, buffer, 0, offset );
            }
        }
    }

    /**
     * Write a single character.  The character to be written is contained in
     * the 16 low-order bits of the given integer value; the 16 high-order bits
     * are ignored.
     *
     * @param c  int specifying a character to be written.
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void write(int c) throws IOException {
        buffer[ offset++ ] = (char) c;
        if( offset == buffer.length ) {
            parentWriter.write( buffer, 0, buffer.length );
            offset = 0;
        }
    }

    /**
     * Write an array of characters.
     *
     * @param  cbuf  Array of characters to be written
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void write(char cbuf[]) throws IOException {
        write(cbuf, 0, cbuf.length);
    }

    /**
     * Write a string.
     *
     * @param  str  String to be written
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void write(String str) throws IOException {
        write(str, 0, str.length());
    }

    /**
     * Write a portion of a string.
     *
     * @param  str  A String
     * @param  off  Offset from which to start writing characters
     * @param  len  Number of characters to write
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void write(String str, int off, int len) throws IOException {
        int available = buffer.length - offset;
        if( len < available ) {
            str.getChars( off, off + len, buffer, offset );
            offset += len;
            return;
        }
        str.getChars( off, off + available, buffer, offset );
        parentWriter.write( buffer, 0, buffer.length );
        if( len - available >= buffer.length ) {
            parentWriter.write( str, off + available, len - available );
            offset = 0;
        } else {
            offset = len - available;
            if( offset > 0 ) {
                str.getChars( off + available, off + len, buffer, 0 );
            }
        }
    }

    /**
     * Flush the stream.  If the stream has saved any characters from the
     * various write() methods in a buffer, write them immediately to their
     * intended destination.  Then, if that destination is another character or
     * byte stream, flush it.  Thus one flush() invocation will flush all the
     * buffers in a chain of Writers and OutputStreams.
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void flush() throws IOException {
        if( offset > 0 ) {
            parentWriter.write( buffer, 0, offset );
            offset = 0;
        }
        parentWriter.flush();
    }

    /**
     * Flush the buffer.  If the buffer has saved any characters from the
     * various write() methods, write them immediately to their
     * intended destination.
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void flushBuffer() throws IOException {
        if( offset > 0 ) {
            parentWriter.write( buffer, 0, offset );
            offset = 0;
        }
    }

    /**
     * Close the stream, flushing it first.  Once a stream has been closed,
     * further write() or flush() invocations will cause an IOException to be
     * thrown.  Closing a previously-closed stream, however, has no effect.
     *
     * @exception  java.io.IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        flush();
        parentWriter.close();
    }
}
