/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.template.instruction;

import freemarker.template.TemplateProcessor;
import freemarker.template.compiler.TemplateBuilder;
import freemarker.template.compiler.ParseException;

/**
 * An interface that parsed instructions must implement.
 *
 * @version $Id: Instruction.java 1096 2005-09-06 13:59:17Z run2000 $
 */
public interface Instruction {

    /**
     * Enumerates the type of end instructions that can terminate a given
     * instruction.
     */
    enum EndType {
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is not an end instruction.
         */
        NONE,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a break instruction.
         */
        BREAK,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a case instruction.
         */
        CASE,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a compress end instruction.
         */
        COMPRESS_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is an else instruction.
         */
        ELSE,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a function end instruction.
         */
        FUNCTION_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is an if end instruction.
         */
        IF_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a list end instruction.
         */
        LIST_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a switch end instruction.
         */
        SWITCH_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a comment end instruction.
         */
        COMMENT_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a noparse end instruction.
         */
        NOPARSE_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a transform end instruction.
         */
        TRANSFORM_END,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is an elseif instruction.
         */
        ELSEIF,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is a default instruction.
         */
        DEFAULT,
        /**
         * Constant returned by <code>getEndType()</code> indicating that this
         * is an assign end instruction.
         */
        ASSIGN_END
    }

    /**
     * Is this an end instruction?
     *
     * @return <code>true</code> if this instruction marks the end of a
     * statement, otherwise <code>false</code>
     */
    public boolean isEndInstruction();

    /**
     * Retrieve the type of end instruction, if any.
     *
     * @return the type of this instruction if it is an end instruction,
     * otherwise <code>NONE</code>.
     */
    public EndType getEndType();

    /**
     * A {@link freemarker.template.compiler.TemplateBuilder} can call
     * this method to have an <code>Instruction</code> call it back to be built.
     * The <code>Instruction</code> will call the appropriate
     * <code>TemplateBuilder.buildStatement()</code>
     * method for its subclass, passing back a reference to itself.
     * This approach is intended to make type-checking of
     * <code>Instruction</code> objects unnecessary.
     *
     * @param builder the builder to be called back by this method
     */
    public TemplateProcessor callBuilder(TemplateBuilder builder)
            throws ParseException;
}
