/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import java.io.*;
import java.util.Properties;

/**
 * <p>Abstract class used to implement test cases for FreeMarker.  This
 * class is an abstract subclass of the <code>TestCase</code> class
 * provided by <a href="http://junit.sourceforge.net/"
 * target="_top">JUnit</a>.  All FM-Classic testcases are subclassed off
 * this class.
 *
 * <p>This class offers functionality to its subclasses to ease testing where
 * a template is read in, and then written out with a specific data model.
 * Passing the test means that the resulting page matches a given reference
 * text. This class also supports the creation of those reference texts.</p>
 *
 * <p>To write a new test by subclassing this class, you'll want to do the
 * following:</p>
 *
 * <ul>
 *
 *  <li> Create a template which your test case will use as input, and store
 *       it in the <code>templates</code> directory located immediately below
 *       this directory.  Say it is called <code>test-foo.html</code>.</li>
 *
 * <li> Create a subclass of <code>AbstractTestCase</code>, and in the
 *       <code>setUp</code> method, call <code>setUpFile("test-foo.html")</code>.
 *       This will initialize several protected instance variables as
 *       follows: <code>m_aReferenceText</code> will contain the contents of
 *       file "reference/test-foo.html", <code>m_aTemplateText</code> will
 *       contain the contents of file "templates/test-foo.html", and
 *       <code>m_aFilename</code> will contain the string "test-foo.html".
 *       You can also set up a data model, as per usual JUnit practice.</li>
 *
 *  <li> Now, in the runTest method of your subclass, create a new Template
 *       from a StringReader constructed around <code>m_aTemplateText</code>.
 *       Process this with your model, and send the output to a
 *       <code>StringWriter</code>.  Finally, call the
 *       <code>showTestResults</code> method, passing in the
 *       <code>m_aReferenceText</code> and the results of your test (from your
 *       <code>StringWriter</code> object).  This will signal an error
 *       if there is a difference.</li>
 *
 *  <li> In addition, if there is a difference, the actual output produced
 *       by the test will be written out in the current directory (wherever
 *       the test was called from), in this case, as "test-foo.html".
 *       This makes it possible to examine the test result and see what
 *       was actually output. It also makes it possible to copy that
 *       into the reference/ directory.</li>
 *
 * </ul>
 *
 * @version $Id: AbstractTestCase.java 1012 2004-10-18 05:47:34Z run2000 $
 */
public abstract class AbstractTestCase extends junit.framework.TestCase {
    private static final int TEST_BUFFER_SIZE = 1024;
    /** The reference text against which the test case output is compared. */
    protected String m_aReferenceText;
    /** The raw, unparsed text of the template. Templates are compiled from this text. */
    protected String m_aTemplateText;
    /** The filename of the reference text if the text doesn't match. */
    protected String m_aFilename;

    /**
     * Creates new AbstractTestCase, with a filename from which to populate
     * reference text and template text.
     */
    public AbstractTestCase( String aTestName ) {
        super( aTestName );
    }

    /**
     * Sets up the reference and template files to be used for the test.
     * This would normally be called from the setUp() method in the JUnit
     * framework.
     *
     * @param aFilename the filename to be used for retrieving the reference
     * text and the template
     */
    protected void setUpFiles( String aFilename ) {
        try {
            m_aTemplateText = getTemplateText( aFilename );
        } catch( IOException e ) {
            throw new RuntimeException( e.getMessage() );
        }

        try {
            m_aReferenceText = getReferenceText( aFilename );
        } catch( IOException e ) {
            m_aReferenceText = "";
        }

        m_aFilename = aFilename;
    }

    /**
     * Reads text from a file. The file is obtained relative to the
     * implementing class.
     *
     * @param aFilename the filename to read from
     * @return the text of the specified file
     */
    protected String getTextFromFile( String aFilename ) throws IOException {
        Class thisClass = this.getClass();
        InputStream cStream;
        Reader cReader;
        StringBuffer cBuffer = new StringBuffer();
        char[] aBuffer = new char[ TEST_BUFFER_SIZE ];
        int nLength;

        cStream = thisClass.getResourceAsStream( aFilename );

        if( cStream == null ) {
            throw new IOException( "Could not find stream " + aFilename );
        }
        cReader = new InputStreamReader( cStream );

        nLength = cReader.read( aBuffer );

        while( nLength > 0 ) {
            cBuffer.append( aBuffer, 0, nLength );
            nLength = cReader.read( aBuffer );
        }
        return cBuffer.toString();
    }

    /**
     * Gets the reference text for the implementing test class.
     *
     * @param aFilename the filename of the text to be retrieved, relative
     * to the <code>reference</code> directory
     * @return the text of the specified file
     */
    protected String getReferenceText( String aFilename ) throws IOException {

        return getTextFromFile( "reference/" + aFilename );
    }

    /**
     * Gets the template text for the implementing test class.
     *
     * @param aFilename the filename of the text to be retrieved, relative
     * to the <code>template</code> directory
     * @return the text of the specified file
     */
    protected String getTemplateText( String aFilename ) throws IOException {

        return getTextFromFile( "template/" + aFilename );
    }

    /**
     * Performs the test on the output text to indicate whether the text is
     * identical to the reference text.
     *
     * @param aReference the reference text to be compared
     * @param aOutput the text that was generated by the test case
     * @throws TestCaseException the text is in some way different
     */
    protected void isTextIdentical( String aReference, String aOutput ) throws TestCaseException {
        int nReferenceLen, nOutputLen, nMinimumLen;
        int nLine, nCharacter, iCurrent;

        nReferenceLen = aReference.length();
        nOutputLen = aOutput.length();

        if( nOutputLen < nReferenceLen ) {
            nMinimumLen = nOutputLen;
        } else {
            nMinimumLen = nReferenceLen;
        }

        nLine = 1;
        nCharacter = 0;

        for( iCurrent = 0; iCurrent < nMinimumLen; iCurrent++ ) {
            if( aReference.charAt( iCurrent ) != aOutput.charAt( iCurrent )) {
                throw new TestCaseException( "Difference encountered at line " +
                    Integer.toString( nLine ) + ", character " +
                    Integer.toString( nCharacter ) + '.' );
            }
            nCharacter++;

            if( aReference.charAt( iCurrent ) == 10 ) {
                nLine++;
                nCharacter = 0;
            }
        }
        if( nOutputLen > nReferenceLen ) {
            throw new TestCaseException( "Output text is " +
                ( Integer.toString( nOutputLen - nReferenceLen )) +
                " characters longer than reference text." );
        } else if( nReferenceLen > nOutputLen ) {
            throw new TestCaseException( "Output text is " +
                ( Integer.toString( nReferenceLen - nOutputLen )) +
                " characters shorter than reference text." );
        }
    }

    /**
     * Verify that the output of a test is identical to the reference text.
     * If they are identical, say nothing (in the JUnit tradition).  If
     * they differ, output the first line number and character where they
     * differ.
     *
     * @param aReference the reference text
     * @param aOutput the output from the test
     */
    protected void showTestResults( String aReference, String aOutput ) {

        try {
            isTextIdentical( m_aReferenceText, aOutput );
        } catch( TestCaseException error ) {
            try {
                writeText( m_aFilename, aOutput );
            } catch( IOException writeException ) {
                fail( writeException.getMessage() );
            }
            assertTrue( error.getMessage(), false );
        }
    }

    /**
     * Writes text to a given filename. Useful when we want to set up the
     * reference file for future tests to compare against.
     *
     * @param aFilename the name of the file to be written
     * @param aText the text to be written to the file
     * @throws IOException the file could not be written
     */
    protected void writeText( String aFilename, String aText ) throws IOException {
        Writer cWriter = new FileWriter( aFilename );

        cWriter.write( aText );
        cWriter.flush();
        cWriter.close();
    }

    /**
     * Retrieve the root path of the FM-Classic distribution from a
     * properties file. This is used to set the root of the FM-Classic
     * cache, which in turn is used to find the example source code.
     *
     * @return the path to the root of the FM-Classic distribution
     * @throws IOException the <code>testcase.properties</code> file could
     * not be read
     */
    protected String getTestcasePath() throws IOException {
        InputStream input = getClass().getResourceAsStream( "testcase.properties" );
        Properties properties = new Properties();

        properties.load( input );
        input.close();

        return (String)properties.get( "testcasePath" );
    }

}
