/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.Test;
import junit.framework.TestSuite;

import java.util.*;

import freemarker.template.*;
import freemarker.ext.beans.*;
import freemarker.testcase.beans.SimpleBean;

/**
 * A simple test case for testing most aspects of the
 * <code>freemarker.ext.beans</code> package.
 *
 * @author Nicholas Cull
 * @version $Id: TestBeanModels.java 1124 2005-10-04 10:48:45Z run2000 $
 */
public final class TestBeanModels extends TestCase {

    /**
     * Constructor that passes the name of the test up to the JUnit
     * constructor.
     *
     * @param aTestname the name of the test
     */
    public TestBeanModels( String aTestname ) {
        super( aTestname );
    }

    /**
     * Performs the bean model test of the List model.
     */
    public void testListModel() throws Exception {

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        List<String> items = new ArrayList<String>();
        items.add( "test1" );
        items.add( "test2" );

        modelRoot.put( "beans", BeansWrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a ListModel", hashModel instanceof ListModel );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodModel );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 2 );
        //System.out.println( numberModel.getAsNumber() );

        TemplateScalarModel scalarModel = (TemplateScalarModel)((TemplateIndexedModel)hashModel).getAtIndex( 0 );
        assertEquals( "test1", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        scalarModel = (TemplateScalarModel)((TemplateIndexedModel)hashModel).getAtIndex( 1 );
        assertEquals( "test2", scalarModel.getAsString() );
        //System.out.println( scalarModel.getAsString() );

        TemplateListModel2 listModel = (TemplateListModel2)hashModel.get( "iterator" );
        assertTrue( "Expecting an IteratorModel", listModel instanceof IteratorModel );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 2, but is " + index, 2, index );
        listModel.releaseIterator( iteratorModel );

        // Again, directly this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 2, but is " + index, 2, index );
        listModel.releaseIterator( iteratorModel );
    }

    /**
     * Performs the bean model test of the Collection model.
     */
    public void testCollectionModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        Collection<String> items = new TreeSet<String>();
        items.add( "test1" );
        items.add( "test2" );
        items.add( "test3" );

        modelRoot.put( "beans", BeansWrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a CollectionModel", hashModel instanceof CollectionModel );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodModel );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 3 );

        TemplateListModel2 listModel = (TemplateListModel2)hashModel.get( "iterator" );
        assertTrue( "Expecting an IteratorModel", listModel instanceof IteratorModel );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        // Again, directly this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );
    }

    /**
     * Performs the bean model test of the Array model.
     */
    public void testArrayModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        String items[] = new String[ 3 ];
        items[ 0 ] = "test1";
        items[ 1 ] = "test2";
        items[ 2 ] = "test3";

        modelRoot.put( "beans", BeansWrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting an ArrayModel", hashModel instanceof ArrayModel );

        TemplateNumberModel numberModel = (TemplateNumberModel)hashModel.get( "length" );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        //System.out.println( numberModel.getClass().toString() );

        assertEquals( numberModel.getAsNumber(), 3 );

        TemplateListModel2 listModel = (TemplateListModel2)hashModel.get( "iterator" );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        assertTrue( "Expecting a FastIndexedIterator", iteratorModel instanceof FastIndexedIterator );
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        // Again, directly this time
        listModel = (TemplateListModel2)hashModel;
        iteratorModel = listModel.templateIterator();
        assertTrue( "Expecting a FastIndexedIterator", iteratorModel instanceof FastIndexedIterator );
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );
    }

    /**
     * Performs the bean model test of the Map model.
     */
    public void testMapModel() throws Exception {

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        Map<String, Long> items = new TreeMap<String, Long>();
        items.put( "test1", new Long( 1 ));
        items.put( "test2", new Long( 2 ));
        items.put( "test3", new Long( 3 ));

        modelRoot.put( "beans", BeansWrapper.wrap( items ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "beans" );
        //System.out.println( hashModel.getClass().toString() );
        assertTrue( "Expecting a MapModel", hashModel instanceof MapModel );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateMethodModel2 methodModel = (TemplateMethodModel2)hashModel.get( "size" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodModel );
        //System.out.println( methodModel.getClass().toString() );
        TemplateNumberModel numberModel = (TemplateNumberModel)methodModel.exec( arguments );

        assertEquals( numberModel.getAsNumber(), 3 );

        arguments.clear();

        List<TemplateModel> arguments2 = new ArrayList<TemplateModel>();

        methodModel = (TemplateMethodModel2)hashModel.get( "keySet" );
        assertTrue( "Expecting a MethodModel", methodModel instanceof MethodModel );
        //System.out.println( methodModel.getClass().toString() );
        TemplateListModel2 listModel = (TemplateListModel2)methodModel.exec( arguments2 );
        //System.out.println( listModel.getClass().toString() );

        TemplateIteratorModel iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        TemplateModel item;
        int index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            //System.out.println( item.getClass().toString() );
            String value = ((TemplateScalarModel)item).getAsString();
            assertEquals( "Item should match index", "test" + index, value );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

        // Now try and get the values from the map
        iteratorModel = listModel.templateIterator();
        //System.out.println( iteratorModel.getClass().toString() );

        index = 0;

        while( iteratorModel.hasNext() ) {
            item = iteratorModel.next();
            index++;
            assertTrue( "Expecting a FastScalar", item instanceof FastScalar );
            arguments2.clear();
            arguments2.add( item );
            TemplateNumberModel result = (TemplateNumberModel)((TemplateMethodModel2)hashModel).exec( arguments2 );
            assertTrue( "Expecting a SimpleNumberModel", result instanceof SimpleNumberModel );
            //System.out.println( result.getClass().toString() );
            assertEquals( "Item should match index", index, result.getAsNumber() );

        }
        assertEquals( "Index should be 3, but is " + index, 3, index );
        listModel.releaseIterator( iteratorModel );

    }

    /**
     * Performs the bean model test of the Root Map model.
     */
    public void testRootMapModel() throws Exception {
        Map<String, TemplateModel> items = new TreeMap<String, TemplateModel>();

        items.put( "test1", new SimpleNumberModel( 1 ));
        items.put( "test2", new SimpleNumberModel( 2 ));
        items.put( "test3", new SimpleNumberModel( 3 ));

        TemplateWriteableHashModel modelRoot = new RootMapModel( items );

        TemplateNumberModel number = (TemplateNumberModel)modelRoot.get( "test1" );
        assertTrue( "Expecting a SimpleNumberModel", number instanceof SimpleNumberModel );
        assertEquals( number.getAsNumber(), 1L );

        number = (TemplateNumberModel)modelRoot.get( "test2" );
        assertTrue( "Expecting a SimpleNumberModel", number instanceof SimpleNumberModel );
        assertEquals( number.getAsNumber(), 2L );

        number = (TemplateNumberModel)modelRoot.get( "test3" );
        assertTrue( "Expecting a SimpleNumberModel", number instanceof SimpleNumberModel );
        assertEquals( number.getAsNumber(), 3L );
    }

    /**
     * Performs the bean model test of the Scalar model.
     */
    public void testScalarModel() throws Exception {
        SimpleBean bean = new SimpleBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", BeansWrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an ObjectModel", hashModel instanceof ObjectModel );
        //System.out.println( hashModel.getClass().toString() );

        TemplateScalarModel scalarModel = (TemplateScalarModel)hashModel.get( "intValue" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), 5 );

        scalarModel = (TemplateScalarModel)hashModel.get( "byteValue" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), 127 );

        scalarModel = (TemplateScalarModel)hashModel.get( "longValue" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), -10 );

        scalarModel = (TemplateScalarModel)hashModel.get( "shortValue" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), 42 );

        scalarModel = (TemplateScalarModel)hashModel.get( "message" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a ScalarModel", scalarModel instanceof ScalarModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( scalarModel.getAsString(), "Hello, world!" );

        scalarModel = (TemplateScalarModel)hashModel.get( "floatObject" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), 3 );
        assertEquals( scalarModel.getAsString(), "3.5" );

        scalarModel = (TemplateScalarModel)hashModel.get( "doubleObject" );
        assertNotNull( scalarModel );
        assertTrue( "Expecting a NumberModel", scalarModel instanceof NumberModel );
        //System.out.println( scalarModel.getClass().toString() );
        assertEquals( ((TemplateNumberModel)scalarModel).getAsNumber(), 3 );
        assertEquals( scalarModel.getAsString(), "3.14" );

        TemplateMethodModel2 dummyModel = (TemplateMethodModel2) hashModel.get( "setShortValue" );
        FastNumber argument1 = new FastNumber( 7 );
        List<TemplateModel> arguments = new ArrayList<TemplateModel>();
        arguments.add( argument1 );
        dummyModel.exec( arguments );
        assertEquals( "Expecting short value assigned to 7", bean.getShortValue(), (short)7 );
    }

    /**
     * Performs the bean model test of the Scalar model.
     */
    public void testStaticModel() throws Exception {
        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "system", StaticModel.create( System.class ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "system" );
        assertTrue( "Expecting a StaticModel", hashModel instanceof StaticModel );
        //System.out.println( hashModel.getClass().toString() );

        TemplateMethodModel2 method = (TemplateMethodModel2)hashModel.get( "currentTimeMillis" );
        assertTrue( "Expecting a MethodModel", method instanceof MethodModel );
        //System.out.println( method.getClass().toString() );

        List<TemplateModel> arguments = new ArrayList<TemplateModel>();

        TemplateNumberModel output = (TemplateNumberModel)method.exec( arguments );
        //System.out.println( output.getClass().toString() );
        assertTrue( "Expecting a SimpleNumberModel", output instanceof SimpleNumberModel );
        //System.out.println( output.getAsNumber() );
    }

    /**
     * Performs tests on coersion of Simple and Fast models into forms usable
     * by the <code>freemarker.ext.beans</code> package.
     */
    public void testMethodCalls() throws Exception {
        SimpleBean bean = new SimpleBean();

        // Step 1: create an object model to wrap
        SimpleHash modelRoot = new SimpleHash();

        modelRoot.put( "bean", BeansWrapper.wrap( bean ));

        // Step 2: assert that the model is correct
        TemplateHashModel hashModel = (TemplateHashModel)modelRoot.get( "bean" );
        assertTrue( "Expecting an ObjectModel", hashModel instanceof ObjectModel );
        //System.out.println( hashModel.getClass().toString() );

        TemplateMethodModel2 method = (TemplateMethodModel2)hashModel.get( "checkList" );
        assertTrue( "Expecting a MethodModel", method instanceof MethodModel );
        //System.out.println( method.getClass().toString() );

        // Step 3: call the method model
        SimpleList list = new SimpleList();
        list.add( "test1" );
        list.add( "test2" );
        list.add( "test3" );

        List<FastScalar> objectList = new ArrayList<FastScalar>();
        objectList.add( new FastScalar( "test1" ));
        objectList.add( new FastScalar( "test2" ));
        objectList.add( new FastScalar( "test3" ));

        FastList<?> list2 = new FastList<FastScalar>( objectList );

        List<TemplateModel> argumentList = new ArrayList<TemplateModel>();
        argumentList.add( list );

        TemplateNumberModel result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a SimpleNumberModel", result instanceof SimpleNumberModel );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Again, with FastList
        argumentList.clear();
        argumentList.add( list2 );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a SimpleNumberModel", result instanceof SimpleNumberModel );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Now with the hash models
        method = (TemplateMethodModel2)hashModel.get( "checkMap" );
        assertTrue( "Expecting a MethodModel", method instanceof MethodModel );
        //System.out.println( method.getClass().toString() );

        SimpleHash hash = new SimpleHash();
        hash.put( "test1", "test1" );
        hash.put( "test2", "test2" );
        hash.put( "test3", "test3" );

        Map<String,FastScalar> objectMap = new HashMap<String,FastScalar>();
        FastScalar fastScalar = new FastScalar( "test1" );
        objectMap.put( "test1", fastScalar );
        fastScalar = new FastScalar( "test2" );
        objectMap.put( "test2", fastScalar );
        fastScalar = new FastScalar( "test3" );
        objectMap.put( "test3", fastScalar );

        FastHash<FastScalar> hash2 = new FastHash<FastScalar>( objectMap );

        // Now call the method
        argumentList.clear();
        argumentList.add( hash );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a SimpleNumberModel", result instanceof SimpleNumberModel );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );

        // Again, with FastHash
        argumentList.clear();
        argumentList.add( hash2 );

        result = (TemplateNumberModel)method.exec( argumentList );
        assertTrue( "Expecting a SimpleNumberModel", result instanceof SimpleNumberModel );
        //System.out.println( result.getClass().toString() );

        assertEquals( 3, result.getAsNumber() );
    }

    /**
     * Return a suite of all the tests in this class by reflection.
     *
     * @return a <code>TestSuite</code> of all the methods starting with the
     * word "test"
     */
    public static Test suite() {
        return new TestSuite(TestBeanModels.class);
    }

    /**
     * Bootstrap for the self-test code.
     */
    public static void main( String[] argc ) throws Exception {
        TestBeanModels cTest = new TestBeanModels( "testBeanModels" );
        TestResult cResults = new TestResult();

        cTest.run( cResults );
    }
}
