/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
package freemarker.testcase;

import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import freemarker.template.CacheListener;
import freemarker.template.FileTemplateCache;
import freemarker.template.CacheEvent;

/**
 * Tests expiry of items from the cache. This happens <em>n</em> update
 * cycles after an item is added to the cache. We should be able to see this
 * in the System.out stream.
 *
 * @version $Id: TestCacheExpiry.java 1003 2004-10-16 14:54:58Z run2000 $
 */
public class TestCacheExpiry implements CacheListener {
    private long startTime;
    private int nRemoved = 0;
    private int nFailed = 0;
    private int nUnavailable = 0;
    private int nUpdated = 0;

    /** Creates new TestCacheExpiry */
    public TestCacheExpiry() {
    }

    /**
     * Performs the test. The test results are sent to <code>System.out</code>.
     */
    public void test() throws Exception {
        FileTemplateCache cCache = new FileTemplateCache(
            getTestcasePath(), 2 );

        startTime = System.currentTimeMillis();

        // Use the default implementation
        cCache.setMaximumAge( 3 );
        cCache.setLoadingPolicy( FileTemplateCache.Strategy.LOAD_ON_DEMAND_WITH_REFRESH );
        cCache.addCacheListener(this);

        cCache.getItem( "template/test-comment.html" );

        Thread.sleep( 5000 );

        cCache.getItem( "template/test-compress.html" );

        Thread.sleep( 5000 );

        cCache.getItem( "template/test-compress.html" );
        cCache.getItem( "template/test-noparse.html" );

        Thread.sleep( 2000 );

        cCache.getItem( "template/test-compress.html" );
        cCache.getItem( "template/test-comment.html" );

        Thread.sleep( 3000 );

        cCache.getItem( "template/test-compress.html" );

        Thread.sleep( 5000 );

        System.out.println( "Stopping auto-update at " + ( System.currentTimeMillis() - startTime ));
        cCache.stopAutoUpdate();

        System.out.println( "Elements updated: " + nUpdated );
        System.out.println( "Cache failures: " + nFailed );
        System.out.println( "Cache unavailable: " + nUnavailable );
        System.out.println( "Elements removed: " + nRemoved );
    }

    /**
     * Called when an element is removed from the cache.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementRemoved(CacheEvent e) {
        nRemoved++;
        System.out.println( "Element removed at " + ( System.currentTimeMillis() - startTime ));
    }

    /**
     * Called when a cache element fails to update.  The event fired
     * contains the <code>Exception</code> object.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementUpdateFailed(CacheEvent e) {
        nFailed++;
        System.out.println( "Element update failed at " + ( System.currentTimeMillis() - startTime ));
    }

    /**
     * Called when a cache cannot be updated.  The event fired
     * contains the <code>Exception</code> object.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void cacheUnavailable(CacheEvent e) {
        nUnavailable++;
        System.out.println( "Cache unavailable at " + ( System.currentTimeMillis() - startTime ));
    }

    /**
     * Called when a cache element updates successfully.
     *
     * @param e the {@link freemarker.template.CacheEvent} fired.
     */
    public void elementUpdated(CacheEvent e) {
        nUpdated++;
        System.out.println( "Cache element updated at " + ( System.currentTimeMillis() - startTime ));
    }

    /**
     * Retrieve the root path of the FM-Classic distribution from a
     * properties file. This is used to set the root of the FM-Classic
     * cache, which in turn is used to find the example source code.
     *
     * @return the path to the root of the FM-Classic distribution.
     */
    protected String getTestcasePath() throws IOException {
        InputStream input = getClass().getResourceAsStream( "testcase.properties" );
        Properties properties = new Properties();

        properties.load( input );
        input.close();

        return (String)properties.get( "testcasePath" );
    }

    /**
     * Main code for running the test.
     */
    public static void main(String args[]) throws Exception {
        TestCacheExpiry cTester = new TestCacheExpiry();

        cTester.test();
    }

}
