/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;
import junit.framework.Test;
import junit.framework.TestSuite;
import junit.framework.TestResult;

import freemarker.template.*;
import freemarker.ext.misc.ExtendedHash;
import freemarker.ext.misc.ExtendedList;

/**
 * A simple test case for performing unit tests on ExtendedHash.
 * These complement the tests performed by the other hash tests
 * by improving code coverage for this model in particular.
 *
 * @author Nicholas Cull
 * @version $Id: TestHashModel2.java 1003 2004-10-16 14:54:58Z run2000 $
 */
public final class TestHashModel2 extends TestCase {

    /**
     * Constructor that passes the name of the test up to the JUnit
     * constructor.
     *
     * @param aTestname the name of the test
     */
    public TestHashModel2( String aTestname ) {
        super( aTestname );
    }

    /**
     * Performs an ExtendedHash test.
     */
    public void testSimpleHash() throws Exception {

        // Step 1: create an object model to test
        ExtendedHash extendedHash = new ExtendedHash();
        extendedHash.put( "test1", new FastScalar( "String1" ));
        extendedHash.put( "test2", "String2" );
        extendedHash.put( "test3", 3 );
        extendedHash.put( "test4", new Long( 4 ));
        extendedHash.put( "test5", true );

        // Step 2: assert that the model is correct
        TemplateScalarModel stringModel = (TemplateScalarModel)extendedHash.get( "test1" );
        assertTrue( "Expecting a FastScalar", stringModel instanceof FastScalar );
        String string = stringModel.getAsString();
        assertEquals( "Value is: " + string, string, "String1" );
        TemplateModel compareModel = new FastScalar( "String1" );
        assertEquals( "FastScalars should be equal", stringModel, compareModel );
        assertEquals( "FastScalar hash codes should match", stringModel.hashCode(), compareModel.hashCode() );

        stringModel = (TemplateScalarModel)extendedHash.get( "test2" );
        assertTrue( "Expecting a FastScalar", stringModel instanceof FastScalar );
        string = stringModel.getAsString();
        assertEquals( "Value is: " + string, string, "String2" );
        compareModel = new FastScalar( "String2" );
        assertEquals( "FastScalars should be equal", stringModel, compareModel );
        assertEquals( "FastScalar hash codes should match", stringModel.hashCode(), compareModel.hashCode() );

        TemplateNumberModel numberModel = (TemplateNumberModel)extendedHash.get( "test3" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        long number = numberModel.getAsNumber();
        assertEquals( "Value is: " + number, 3, number );
        compareModel = new FastNumber( 3 );
        assertEquals( "FastNumbers should be equal", numberModel, compareModel );
        assertEquals( "FastNumber hash codes should match", numberModel.hashCode(), compareModel.hashCode() );

        numberModel = (TemplateNumberModel)extendedHash.get( "test4" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        number = numberModel.getAsNumber();
        assertEquals( "Value is: " + number, 4, number );
        compareModel = new FastNumber( 4 );
        assertEquals( "FastNumbers should be equal", numberModel, compareModel );
        assertEquals( "FastNumber hash codes should match", numberModel.hashCode(), compareModel.hashCode() );

        stringModel = (TemplateScalarModel)extendedHash.get( "test5" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastBoolean", stringModel instanceof FastBoolean );
        assertTrue( "FastBoolean should be true", ! stringModel.isEmpty() );
        assertEquals( "FastBooleans should be equal", stringModel, FastBoolean.TRUE );
        assertEquals( "FastBooleans hash codes should match", stringModel.hashCode(), FastBoolean.TRUE.hashCode() );
        assertTrue( "FastBooleans should be the same object", stringModel == FastBoolean.TRUE );
    }

    /**
     * Performs an ExtendedHash test.
     */
    public void testSimpleHash2() throws Exception {

        // Step 1: create an object model to test
        Map<String, TemplateModel> map = new HashMap<String, TemplateModel>();

        map.put( "test1", new FastScalar( "String1" ));
        map.put( "test2", new FastScalar( "String2" ));
        map.put( "test3", new FastNumber( 3 ));
        map.put( "test4", new FastNumber( new Long( 4 )));
        map.put( "test5", FastBoolean.getInstance( true ));

        ExtendedHash extendedHash = new ExtendedHash( map );

        // Step 2: assert that the model is correct
        TemplateScalarModel stringModel = (TemplateScalarModel)extendedHash.get( "test1" );
        assertTrue( "Expecting a FastScalar", stringModel instanceof FastScalar );
        String string = stringModel.getAsString();
        assertEquals( "Value is: " + string, string, "String1" );
        TemplateModel compareModel = new FastScalar( "String1" );
        assertEquals( "FastScalars should be equal", stringModel, compareModel );
        assertEquals( "FastScalar hash codes should match", stringModel.hashCode(), compareModel.hashCode() );

        stringModel = (TemplateScalarModel)extendedHash.get( "test2" );
        assertTrue( "Expecting a FastScalar", stringModel instanceof FastScalar );
        string = stringModel.getAsString();
        assertEquals( "Value is: " + string, string, "String2" );
        compareModel = new FastScalar( "String2" );
        assertEquals( "FastScalars should be equal", stringModel, compareModel );
        assertEquals( "FastScalar hash codes should match", stringModel.hashCode(), compareModel.hashCode() );

        TemplateNumberModel numberModel = (TemplateNumberModel)extendedHash.get( "test3" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        long number = numberModel.getAsNumber();
        assertEquals( "Value is: " + number, 3, number );
        compareModel = new FastNumber( 3 );
        assertEquals( "FastNumbers should be equal", numberModel, compareModel );
        assertEquals( "FastNumber hash codes should match", numberModel.hashCode(), compareModel.hashCode() );

        numberModel = (TemplateNumberModel)extendedHash.get( "test4" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastNumber", numberModel instanceof FastNumber );
        number = numberModel.getAsNumber();
        assertEquals( "Value is: " + number, 4, number );
        compareModel = new FastNumber( 4 );
        assertEquals( "FastNumbers should be equal", numberModel, compareModel );
        assertEquals( "FastNumber hash codes should match", numberModel.hashCode(), compareModel.hashCode() );

        stringModel = (TemplateScalarModel)extendedHash.get( "test5" );
        assertNotNull( numberModel );
        assertTrue( "Expecting a FastBoolean", stringModel instanceof FastBoolean );
        assertTrue( "FastBoolean should be true", ! stringModel.isEmpty() );
        assertEquals( "FastBooleans should be equal", stringModel, FastBoolean.TRUE );
        assertEquals( "FastBooleans hash codes should match", stringModel.hashCode(), FastBoolean.TRUE.hashCode() );
        assertTrue( "FastBooleans should be the same object", stringModel == FastBoolean.TRUE );
    }

    /**
     * Performs an ExtendedHash test.
     */
    public void testAddRemove() throws Exception {

        // Step 1: create an object model to test
        ExtendedHash extendedHash = new ExtendedHash();

        // Step 2: assert that the model is correct
        assertTrue( "Hash should be empty", extendedHash.isEmpty() );

        extendedHash.put( "test1", "String1" );
        extendedHash.put( "test2", 2 );
        extendedHash.put( "test3", true );

        assertTrue( "Hash should not be empty", ! extendedHash.isEmpty() );

        extendedHash.remove( "test1" );
        assertTrue( "Hash should not be empty", ! extendedHash.isEmpty() );
        extendedHash.remove( "test2" );
        assertTrue( "Hash should not be empty", ! extendedHash.isEmpty() );
        extendedHash.remove( "test3" );
        assertTrue( "Hash should be empty", extendedHash.isEmpty() );
    }

    /**
     * Performs an ExtendedHash test of the special keys.
     */
    public void testSpecialKeys() throws Exception {

        // Step 1: create an object model to test
        ExtendedHash extendedHash = new ExtendedHash();

        // Step 2: assert that the model is correct
        assertTrue( "Hash should be empty", extendedHash.isEmpty() );

        extendedHash.put( "test1", "String1" );

        assertTrue( "Hash should not be empty", ! extendedHash.isEmpty() );

        TemplateIndexedModel listResult = (TemplateIndexedModel)extendedHash.get( "_keys" );
        assertTrue( "Result should be a ExtendedList", listResult instanceof ExtendedList );

        TemplateScalarModel scalarResult = (TemplateScalarModel)listResult.getAtIndex( 0 );
        assertEquals( "Scalar model should have value 'test1'", scalarResult.getAsString(), "test1" );

        listResult = (TemplateIndexedModel)extendedHash.get( "_values" );
        assertTrue( "Result should be a ExtendedList", listResult instanceof ExtendedList );

        scalarResult = (TemplateScalarModel)listResult.getAtIndex( 0 );
        assertEquals( "Scalar model should have value 'String1'", scalarResult.getAsString(), "String1" );

        TemplateNumberModel numberResult = (TemplateNumberModel)extendedHash.get( "_size" );
        assertTrue( "Result should be a FastNumber", numberResult instanceof FastNumber );
        assertEquals( "Number model should have value 1", numberResult.getAsNumber(), 1 );

        scalarResult = (TemplateScalarModel)extendedHash.get( "test1" );
        assertTrue( "Result should be a FastScalar", scalarResult instanceof FastScalar );
        assertEquals( "Result should be value 'String1", scalarResult.getAsString(), "String1" );

    }

    /**
     * Return a suite of all the tests in this class by reflection.
     *
     * @return a <code>TestSuite</code> of all the methods starting with the
     * word "test"
     */
    public static Test suite() {
        return new TestSuite(TestHashModel2.class);
    }

    /**
     * Bootstrap for the self-test code.
     */
    public static void main( String[] argc ) throws Exception {
        TestHashModel2 cTest = new TestHashModel2( "testHashModel" );
        TestResult cResults = new TestResult();

        cTest.run( cResults );
    }
}
