/*
 * FreeMarker: a tool that allows Java programs to generate HTML
 * output using templates.
 * Copyright (C) 1998-2004 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

package freemarker.testcase;

import freemarker.template.*;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.io.StringReader;

import junit.framework.*;

/**
 * A second test class for testing FM-Classic's &lt;include&gt; tag. In
 * particular, test whether we can escape from a cache directory's sandbox.
 *
 * @version $Id: TestInclude2.java 1128 2005-10-04 11:36:12Z run2000 $
 * @see TestInclude
 */
public final class TestInclude2 extends AbstractTestCase implements CacheListener {
    private FastHash modelRoot;

    /**
     * Constructor that passes the name of the test up to the JUnit
     * constructor.
     *
     * @param aTestname the name of the test
     */
    public TestInclude2( String aTestname ) {
        super( aTestname );
    }

    /**
     * Set up the test case prior to running.
     */
    public void setUp() {
        setUpFiles( "test-include2.html" );

        // Make a template data model.
        modelRoot = new FastHash();
        modelRoot.put( "message", new FastScalar( "Hello, world!" ));

    }

    /**
     * Performs the test. A fail result will generate information about
     * the failure and pass it up to JUnit.
     */
    public void runTest() throws IOException {
        FileTemplateCache cCache;
        Template cTemplate;
        StringWriter cWriter = new StringWriter();
        String aDirectory;

        cCache = new FileTemplateCache();

        //cCache.setLoadingPolicy( freemarker.template.FileTemplateCache.PRELOAD );
        aDirectory = getTestcasePath() + "/template";
        cCache.setDirectory( new File( aDirectory ));
        cCache.addCacheListener( this );

        cTemplate = new Template( new StringInputSource( m_aTemplateText ));
        cTemplate.setCache( cCache );
        //System.out.println( cTemplate.toString() );

        // Process the template.
        cTemplate.process( modelRoot, cWriter );
        cCache.stopAutoUpdate();
        cCache.removeCacheListener( this );

        showTestResults( m_aReferenceText, cWriter.toString() );
    }

    /**
     * Called when a cache cannot be updated. The event fired
     * contains the <code>Exception</code> object.
     *
     * @param event the {@link CacheEvent} fired.
     */
    public void cacheUnavailable(CacheEvent event) {
        fail( "Cache is not available" );
    }

    /**
     * Called when a cache element updates successfully.
     *
     * @param e the {@link CacheEvent} fired.
     */
    public void elementUpdated(CacheEvent e) {
        assertNotNull( e );
        assertNotNull( e.getElementName() );
    }

    /**
     * Called when an element is removed from the cache.
     *
     * @param e the {@link CacheEvent} fired.
     */
    public void elementRemoved(CacheEvent e) {
        assertNotNull( e );
        assertNotNull( e.getElementName() );
    }

    /**
     * Called when a cache element fails to update.  The event fired
     * contains the <code>Exception</code> object.
     *
     * @param e the {@link CacheEvent} fired.
     */
    public void elementUpdateFailed(CacheEvent e) {
        assertNotNull( e );
        assertNotNull( e.getException() );
        assertNotNull( e.getElementName() );
    }

    /**
     * Bootstrap for the self-test code.
     */
    public static void main( String[] argc ) throws Exception {
        AbstractTestCase cTest = new TestInclude2( "test-include2.html" );
        TestResult cResults = new TestResult();

        cTest.run( cResults );
    }
}